# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import logging

from openstack import exceptions as sdk_exc

from metalsmith import _utils
from metalsmith import exceptions


LOG = logging.getLogger(__name__)


class NICs(object):
    """Requested NICs."""

    def __init__(self, connection, node, nics):
        if nics is None:
            nics = []

        if not isinstance(nics, collections.Sequence):
            raise TypeError("NICs must be a list of dicts")

        for nic in nics:
            if not isinstance(nic, collections.Mapping):
                raise TypeError("Each NIC must be a dict got %s" % nic)

        self._node = node
        self._connection = connection
        self._nics = nics
        self._validated = None
        self.created_ports = []
        self.attached_ports = []

    def validate(self):
        """Validate provided NIC records."""
        if self._validated is not None:
            return

        result = []
        for nic in self._nics:
            if 'port' in nic:
                result.append(('port', self._get_port(nic)))
            elif 'network' in nic:
                result.append(('network', self._get_network(nic)))
            elif 'subnet' in nic:
                result.append(('subnet', self._get_subnet(nic)))
            else:
                raise exceptions.InvalidNIC(
                    'Unknown NIC record type, export "port", "subnet" or '
                    '"network", got %s' % nic)

        self._validated = result

    def create_and_attach_ports(self):
        """Attach ports to the node, creating them if requested."""
        self.validate()

        for nic_type, nic in self._validated:
            if nic_type != 'port':
                port = self._connection.network.create_port(**nic)
                self.created_ports.append(port.id)
                LOG.info('Created port %(port)s for node %(node)s with '
                         '%(nic)s', {'port': _utils.log_res(port),
                                     'node': _utils.log_res(self._node),
                                     'nic': nic})
            else:
                port = nic

            self._connection.baremetal.attach_vif_to_node(self._node,
                                                          port.id)
            LOG.info('Attached port %(port)s to node %(node)s',
                     {'port': _utils.log_res(port),
                      'node': _utils.log_res(self._node)})
            self.attached_ports.append(port.id)

    def detach_and_delete_ports(self):
        """Detach attached port and delete previously created ones."""
        detach_and_delete_ports(self._connection, self._node,
                                self.created_ports, self.attached_ports)

    def _get_port(self, nic):
        """Validate and get the NIC information for a port.

        :param nic: NIC information in the form ``{"port": "<port ident>"}``.
        :returns: `Port` object to use.
        """
        unexpected = set(nic) - {'port'}
        if unexpected:
            raise exceptions.InvalidNIC(
                'Unexpected fields for a port: %s' % ', '.join(unexpected))

        try:
            port = self._connection.network.find_port(
                nic['port'], ignore_missing=False)
        except sdk_exc.SDKException as exc:
            raise exceptions.InvalidNIC(
                'Cannot find port %(port)s: %(error)s' %
                {'port': nic['port'], 'error': exc})

        return port

    def _get_network(self, nic):
        """Validate and get the NIC information for a network.

        :param nic: NIC information in the form ``{"network": "<net ident>"}``
            or ``{"network": "<net ident>", "fixed_ip": "<desired IP>"}``.
        :returns: keyword arguments to use when creating a port.
        """
        unexpected = set(nic) - {'network', 'fixed_ip'}
        if unexpected:
            raise exceptions.InvalidNIC(
                'Unexpected fields for a network: %s' % ', '.join(unexpected))

        try:
            network = self._connection.network.find_network(
                nic['network'], ignore_missing=False)
        except sdk_exc.SDKException as exc:
            raise exceptions.InvalidNIC(
                'Cannot find network %(net)s: %(error)s' %
                {'net': nic['network'], 'error': exc})

        port_args = {'network_id': network.id}
        if nic.get('fixed_ip'):
            port_args['fixed_ips'] = [{'ip_address': nic['fixed_ip']}]

        return port_args

    def _get_subnet(self, nic):
        """Validate and get the NIC information for a subnet.

        :param nic: NIC information in the form ``{"subnet": "<id or name>"}``.
        :returns: keyword arguments to use when creating a port.
        """
        unexpected = set(nic) - {'subnet'}
        if unexpected:
            raise exceptions.InvalidNIC(
                'Unexpected fields for a subnet: %s' % ', '.join(unexpected))

        try:
            subnet = self._connection.network.find_subnet(
                nic['subnet'], ignore_missing=False)
        except sdk_exc.SDKException as exc:
            raise exceptions.InvalidNIC(
                'Cannot find subnet %(sub)s: %(error)s' %
                {'sub': nic['subnet'], 'error': exc})

        try:
            network = self._connection.network.get_network(subnet.network_id)
        except sdk_exc.SDKException as exc:
            raise exceptions.InvalidNIC(
                'Cannot find network %(net)s for subnet %(sub)s: %(error)s' %
                {'net': subnet.network_id, 'sub': nic['subnet'], 'error': exc})

        return {'network_id': network.id,
                'fixed_ips': [{'subnet_id': subnet.id}]}


def detach_and_delete_ports(connection, node, created_ports, attached_ports):
    """Detach attached port and delete previously created ones.

    :param connection: `openstacksdk.Connection` instance.
    :param node: `Node` object to detach ports from.
    :param created_ports: List of IDs of previously created ports.
    :param attached_ports: List of IDs of previously attached_ports.
    """
    for port_id in set(attached_ports + created_ports):
        LOG.debug('Detaching port %(port)s from node %(node)s',
                  {'port': port_id, 'node': _utils.log_res(node)})
        try:
            connection.baremetal.detach_vif_from_node(node, port_id)
        except Exception as exc:
            LOG.debug('Failed to remove VIF %(vif)s from node %(node)s, '
                      'assuming already removed: %(exc)s',
                      {'vif': port_id, 'node': _utils.log_res(node),
                       'exc': exc})

    for port_id in created_ports:
        LOG.debug('Deleting port %s', port_id)
        try:
            connection.network.delete_port(port_id, ignore_missing=False)
        except Exception as exc:
            LOG.warning('Failed to delete neutron port %(port)s: %(exc)s',
                        {'port': port_id, 'exc': exc})
        else:
            LOG.info('Deleted port %(port)s for node %(node)s',
                     {'port': port_id, 'node': _utils.log_res(node)})



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import random
import sys
import warnings

from openstack import connection
import six

from metalsmith import _config
from metalsmith import _instance
from metalsmith import _nics
from metalsmith import _scheduler
from metalsmith import _utils
from metalsmith import exceptions
from metalsmith import sources


LOG = logging.getLogger(__name__)

_CREATED_PORTS = 'metalsmith_created_ports'
_ATTACHED_PORTS = 'metalsmith_attached_ports'
_PRESERVE_INSTANCE_INFO_KEYS = {'capabilities', 'traits'}


class Provisioner(_utils.GetNodeMixin):
    """API to deploy/undeploy nodes with OpenStack.

    :param session: `Session` object (from ``keystoneauth``) to use when
        making API requests. Mutually exclusive with **cloud_region**.
    :param cloud_region: cloud configuration object (from ``openstacksdk``)
        to use when making API requests. Mutually exclusive with **session**.
    :param dry_run: boolean value, set to ``True`` to prevent any API calls
        from being actually made.

    :ivar connection: `openstacksdk` `Connection` object used for accessing
        OpenStack API during provisioning.
    """

    def __init__(self, session=None, cloud_region=None, dry_run=False):
        if cloud_region is None:
            if session is None:
                raise TypeError('Either session or cloud_region must '
                                'be provided')
            self.connection = connection.Connection(session=session)
        elif session is not None:
            raise TypeError('Either session or cloud_region must be provided, '
                            'but not both')
        else:
            self.connection = connection.Connection(config=cloud_region)

        self._dry_run = dry_run

    def reserve_node(self, resource_class=None, conductor_group=None,
                     capabilities=None, traits=None, candidates=None,
                     predicate=None):
        """Find and reserve a suitable node.

        Example::

         node = provisioner.reserve_node("compute",
                                         capabilities={"boot_mode": "uefi"})

        :param resource_class: Requested resource class.
        :param conductor_group: Conductor group to pick the nodes from.
            Value ``None`` means any group, use empty string "" for nodes
            from the default group.
        :param capabilities: Requested capabilities as a dict.
        :param traits: Requested traits as a list of strings.
        :param candidates: List of nodes (UUIDs, names or `Node` objects)
            to pick from. The filters (for resource class and capabilities)
            are still applied to the provided list. The order in which
            the nodes are considered is retained.
        :param predicate: Custom predicate to run on nodes. A callable that
            accepts a node and returns ``True`` if it should be included,
            ``False`` otherwise. Any exceptions are propagated to the caller.
        :return: reserved `Node` object.
        :raises: :py:class:`metalsmith.exceptions.ReservationFailed`
        """
        capabilities = capabilities or {}
        if resource_class is None:
            warnings.warn("Not providing resource_class is deprecated as it's "
                          "not compatible with the proposed allocation API",
                          DeprecationWarning)

        if candidates:
            nodes = [self._get_node(node) for node in candidates]
        else:
            kwargs = {}
            if conductor_group:
                kwargs['conductor_group'] = conductor_group
            nodes = list(self.connection.baremetal.nodes(
                associated=False,
                provision_state='available',
                maintenance=False,
                resource_class=resource_class,
                details=True,
                **kwargs))
            if not nodes:
                raise exceptions.NodesNotFound(resource_class, conductor_group)
            # Ensure parallel executions don't try nodes in the same sequence
            random.shuffle(nodes)

        LOG.debug('Candidate nodes: %s', nodes)

        filters = [
            _scheduler.NodeTypeFilter(resource_class, conductor_group),
            _scheduler.CapabilitiesFilter(capabilities),
            _scheduler.TraitsFilter(traits),
        ]
        if predicate is not None:
            filters.append(_scheduler.CustomPredicateFilter(predicate))

        instance_info = {}
        if capabilities:
            instance_info['capabilities'] = capabilities
        if traits:
            instance_info['traits'] = traits
        reserver = _scheduler.IronicReserver(self.connection,
                                             instance_info)

        node = _scheduler.schedule_node(nodes, filters, reserver,
                                        dry_run=self._dry_run)
        LOG.debug('Reserved node: %s', node)
        return node

    def _check_node_for_deploy(self, node):
        """Check that node is ready and reserve it if needed.

        These checks are done outside of the try..except block in
        ``provision_node``, so that we don't touch nodes that fail it at all.
        Particularly, we don't want to try clean up nodes that were not
        reserved by us or are in maintenance mode.
        """
        try:
            node = self._get_node(node)
        except Exception as exc:
            raise exceptions.InvalidNode('Cannot find node %(node)s: %(exc)s' %
                                         {'node': node, 'exc': exc})

        if not node.instance_id:
            if not self._dry_run:
                LOG.debug('Node %s not reserved yet, reserving',
                          _utils.log_res(node))
                self.connection.baremetal.update_node(
                    node, instance_id=node.id)
        elif node.instance_id != node.id:
            raise exceptions.InvalidNode('Node %(node)s already reserved '
                                         'by instance %(inst)s outside of '
                                         'metalsmith, cannot deploy on it' %
                                         {'node': _utils.log_res(node),
                                          'inst': node.instance_id})

        if node.is_maintenance:
            raise exceptions.InvalidNode('Refusing to deploy on node %(node)s '
                                         'which is in maintenance mode due to '
                                         '%(reason)s' %
                                         {'node': _utils.log_res(node),
                                          'reason': node.maintenance_reason})

        return node

    def _check_hostname(self, node, hostname):
        """Check the provided host name.

        If the ``hostname`` is not provided, use either the name or the UUID,
        whichever is appropriate for a host name.

        :return: appropriate hostname
        :raises: ValueError on inappropriate value of ``hostname``
        """
        if hostname is None:
            if node.name and _utils.is_hostname_safe(node.name):
                return node.name
            else:
                return node.id

        if not _utils.is_hostname_safe(hostname):
            raise ValueError("%s cannot be used as a hostname" % hostname)

        existing = self._find_node_by_hostname(hostname)
        if existing is not None and existing.id != node.id:
            raise ValueError("The following node already uses hostname "
                             "%(host)s: %(node)s" %
                             {'host': hostname,
                              'node': _utils.log_res(existing)})

        return hostname

    def provision_node(self, node, image, nics=None, root_size_gb=None,
                       swap_size_mb=None, config=None, hostname=None,
                       netboot=False, capabilities=None, traits=None,
                       wait=None):
        """Provision the node with the given image.

        Example::

         provisioner.provision_node("compute-1", "centos",
                                    nics=[{"network": "private"},
                                          {"network": "external"}],
                                    root_size_gb=50,
                                    wait=3600)

        :param node: Node object, UUID or name. Will be reserved first, if
            not reserved already. Must be in the "available" state with
            maintenance mode off.
        :param image: Image source - one of :mod:`~metalsmith.sources`,
            `Image` name or UUID.
        :param nics: List of virtual NICs to attach to physical ports.
            Each item is a dict with a key describing the type of the NIC:

            * ``{"port": "<port name or ID>"}`` to use the provided pre-created
              port.
            * ``{"network": "<network name or ID>"}`` to create a port on the
              provided network. Optionally, a ``fixed_ip`` argument can be used
              to specify an IP address.
            * ``{"subnet": "<subnet name or ID>"}`` to create a port with an IP
              address from the provided subnet. The network is determined from
              the subnet.

        :param root_size_gb: The size of the root partition. By default
            the value of the local_gb property is used.
        :param swap_size_mb: The size of the swap partition. It's an error
            to specify it for a whole disk image.
        :param config: :py:class:`metalsmith.InstanceConfig` object with
            the configuration to pass to the instance.
        :param hostname: Hostname to assign to the instance. Defaults to the
            node's name or UUID.
        :param netboot: Whether to use networking boot for final instances.
        :param capabilities: Requested capabilities of the node. If present,
            overwrites the capabilities set by :meth:`reserve_node`.
            Note that the capabilities are not checked against the ones
            provided by the node - use :meth:`reserve_node` for that.
        :param traits: Requested traits of the node. If present, overwrites
            the traits set by :meth:`reserve_node`. Note that the traits are
            not checked against the ones provided by the node - use
            :meth:`reserve_node` for that.
        :param wait: How many seconds to wait for the deployment to finish,
            None to return immediately.
        :return: :py:class:`metalsmith.Instance` object with the current
            status of provisioning. If ``wait`` is not ``None``, provisioning
            is already finished.
        :raises: :py:class:`metalsmith.exceptions.Error`
        """
        if config is None:
            config = _config.InstanceConfig()
        if isinstance(image, six.string_types):
            image = sources.GlanceImage(image)

        node = self._check_node_for_deploy(node)
        nics = _nics.NICs(self.connection, node, nics)

        try:
            hostname = self._check_hostname(node, hostname)
            root_size_gb = _utils.get_root_disk(root_size_gb, node)

            image._validate(self.connection)

            nics.validate()

            if capabilities is None:
                capabilities = node.instance_info.get('capabilities') or {}

            if self._dry_run:
                LOG.warning('Dry run, not provisioning node %s',
                            _utils.log_res(node))
                return node

            nics.create_and_attach_ports()

            capabilities['boot_option'] = 'netboot' if netboot else 'local'

            instance_info = self._clean_instance_info(node.instance_info)
            instance_info['root_gb'] = root_size_gb
            instance_info['capabilities'] = capabilities
            instance_info[self.HOSTNAME_FIELD] = hostname
            extra = node.extra.copy()
            extra[_CREATED_PORTS] = nics.created_ports
            extra[_ATTACHED_PORTS] = nics.attached_ports
            instance_info.update(image._node_updates(self.connection))
            if traits is not None:
                instance_info['traits'] = traits
            if swap_size_mb is not None:
                instance_info['swap_mb'] = swap_size_mb

            LOG.debug('Updating node %(node)s with instance info %(iinfo)s '
                      'and extras %(extra)s', {'node': _utils.log_res(node),
                                               'iinfo': instance_info,
                                               'extra': extra})
            node = self.connection.baremetal.update_node(
                node, instance_info=instance_info, extra=extra)
            self.connection.baremetal.validate_node(node)

            LOG.debug('Generating a configdrive for node %s',
                      _utils.log_res(node))
            cd = config.build_configdrive(node, hostname)
            # TODO(dtantsur): move this to openstacksdk?
            if not isinstance(cd, six.string_types):
                cd = cd.decode('utf-8')
            LOG.debug('Starting provisioning of node %s', _utils.log_res(node))
            self.connection.baremetal.set_node_provision_state(
                node, 'active', config_drive=cd)
        except Exception:
            exc_info = sys.exc_info()

            try:
                LOG.error('Deploy attempt failed on node %s, cleaning up',
                          _utils.log_res(node))
                self._clean_up(node, nics=nics)
            except Exception:
                LOG.exception('Clean up failed')

            six.reraise(*exc_info)

        LOG.info('Provisioning started on node %s', _utils.log_res(node))

        if wait is not None:
            LOG.debug('Waiting for node %(node)s to reach state active '
                      'with timeout %(timeout)s',
                      {'node': _utils.log_res(node), 'timeout': wait})
            instance = self.wait_for_provisioning([node], timeout=wait)[0]
            LOG.info('Deploy succeeded on node %s', _utils.log_res(node))
        else:
            # Update the node to return it's latest state
            node = self._get_node(node, refresh=True)
            instance = _instance.Instance(self.connection, node)

        return instance

    def wait_for_provisioning(self, nodes, timeout=None, delay=None):
        """Wait for nodes to be provisioned.

        Loops until all nodes finish provisioning.

        :param nodes: List of nodes (UUID, name, `Node` object or
            :py:class:`metalsmith.Instance`).
        :param timeout: How much time (in seconds) to wait for all nodes
            to finish provisioning. If ``None`` (the default), wait forever
            (more precisely, until the operation times out on server side).
        :param delay: DEPRECATED, do not use.
        :return: List of updated :py:class:`metalsmith.Instance` objects if
            all succeeded.
        :raises: :py:class:`metalsmith.exceptions.DeploymentFailure`
            if the deployment failed or timed out for any nodes.
        """
        if delay is not None:
            warnings.warn("The delay argument to wait_for_provisioning is "
                          "deprecated and has not effect", DeprecationWarning)
        nodes = [self._get_node(n, accept_hostname=True) for n in nodes]
        nodes = self.connection.baremetal.wait_for_nodes_provision_state(
            nodes, 'active', timeout=timeout)
        return [_instance.Instance(self.connection, node) for node in nodes]

    def _clean_instance_info(self, instance_info):
        return {key: value
                for key, value in instance_info.items()
                if key in _PRESERVE_INSTANCE_INFO_KEYS}

    def _clean_up(self, node, nics=None):
        if nics is None:
            created_ports = node.extra.get(_CREATED_PORTS, [])
            attached_ports = node.extra.get(_ATTACHED_PORTS, [])
            _nics.detach_and_delete_ports(self.connection, node,
                                          created_ports, attached_ports)
        else:
            nics.detach_and_delete_ports()

        extra = node.extra.copy()
        for item in (_CREATED_PORTS, _ATTACHED_PORTS):
            extra.pop(item, None)
        LOG.debug('Updating node %(node)s with empty instance info (was '
                  '%(iinfo)s) and extras %(extra)s and releasing the lock',
                  {'node': _utils.log_res(node),
                   'iinfo': node.instance_info,
                   'extra': extra})
        try:
            self.connection.baremetal.update_node(
                node, instance_info={}, extra=extra, instance_id=None)
        except Exception as exc:
            LOG.debug('Failed to clear node %(node)s extra: %(exc)s',
                      {'node': _utils.log_res(node), 'exc': exc})

    def unprovision_node(self, node, wait=None):
        """Unprovision a previously provisioned node.

        :param node: `Node` object, :py:class:`metalsmith.Instance`,
            hostname, UUID or node name.
        :param wait: How many seconds to wait for the process to finish,
            None to return immediately.
        :return: the latest `Node` object.
        """
        node = self._get_node(node, accept_hostname=True)
        if self._dry_run:
            LOG.warning("Dry run, not unprovisioning")
            return

        self._clean_up(node)
        node = self.connection.baremetal.set_node_provision_state(
            node, 'deleted', wait=False)

        LOG.info('Deleting started for node %s', _utils.log_res(node))

        if wait is not None:
            node = self.connection.baremetal.wait_for_nodes_provision_state(
                [node], 'available', timeout=wait)[0]
            LOG.info('Node %s undeployed successfully', _utils.log_res(node))

        return node

    def show_instance(self, instance_id):
        """Show information about instance.

        :param instance_id: hostname, UUID or node name.
        :return: :py:class:`metalsmith.Instance` object.
        :raises: :py:class:`metalsmith.exceptions.InvalidInstance`
            if the instance is not a valid instance.
        """
        return self.show_instances([instance_id])[0]

    def show_instances(self, instances):
        """Show information about instance.

        More efficient than calling :meth:`show_instance` in a loop, because
        it caches the node list.

        :param instances: list of hostnames, UUIDs or node names.
        :return: list of :py:class:`metalsmith.Instance` objects in the same
            order as ``instances``.
        :raises: :py:class:`metalsmith.exceptions.InvalidInstance`
            if one of the instances are not valid instances.
        """
        with self._cache_node_list_for_lookup():
            result = [
                _instance.Instance(
                    self.connection,
                    self._get_node(inst, accept_hostname=True))
                for inst in instances
            ]
        # NOTE(dtantsur): do not accept node names as valid instances if they
        # are not deployed or being deployed.
        missing = [inst for (res, inst) in zip(result, instances)
                   if res.state == _instance.InstanceState.UNKNOWN]
        if missing:
            raise exceptions.InvalidInstance(
                "Node(s)/instance(s) %s are not valid instances"
                % ', '.join(map(str, missing)))
        return result

    def list_instances(self):
        """List instances deployed by metalsmith.

        :return: list of :py:class:`metalsmith.Instance` objects.
        """
        nodes = self.connection.baremetal.nodes(associated=True, details=True)
        instances = [i for i in
                     (_instance.Instance(self.connection, node)
                      for node in nodes)
                     if i.state != _instance.InstanceState.UNKNOWN]
        return instances



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import mock
from openstack.baremetal import configdrive
import testtools

from metalsmith import _config


class TestInstanceConfig(testtools.TestCase):
    def setUp(self):
        super(TestInstanceConfig, self).setUp()
        self.node = mock.Mock(id='1234')
        self.node.name = 'node name'

    def _check(self, config, expected_metadata, expected_userdata=None):
        expected_m = {'public_keys': {},
                      'uuid': '1234',
                      'name': 'node name',
                      'hostname': 'example.com',
                      'launch_index': 0,
                      'availability_zone': '',
                      'files': [],
                      'meta': {}}
        expected_m.update(expected_metadata)

        with mock.patch.object(configdrive, 'build', autospec=True) as mb:
            result = config.build_configdrive(self.node, "example.com")
            mb.assert_called_once_with(expected_m, mock.ANY)
            self.assertIs(result, mb.return_value)
            user_data = mb.call_args[0][1]

        if expected_userdata:
            self.assertIsNotNone(user_data)
            user_data = user_data.decode('utf-8')
            header, user_data = user_data.split('\n', 1)
            self.assertEqual('#cloud-config', header)
            user_data = json.loads(user_data)
        self.assertEqual(expected_userdata, user_data)

    def test_default(self):
        config = _config.InstanceConfig()
        self._check(config, {})

    def test_ssh_keys(self):
        config = _config.InstanceConfig(ssh_keys=['abc', 'def'])
        self._check(config, {'public_keys': {'0': 'abc', '1': 'def'}})

    def test_ssh_keys_as_dict(self):
        config = _config.InstanceConfig(ssh_keys={'default': 'abc'})
        self._check(config, {'public_keys': {'default': 'abc'}})

    def test_add_user(self):
        config = _config.InstanceConfig()
        config.add_user('admin')
        self._check(config, {},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel']}]})

    def test_add_user_admin(self):
        config = _config.InstanceConfig()
        config.add_user('admin', admin=False)
        self._check(config, {},
                    {'users': [{'name': 'admin'}]})

    def test_add_user_sudo(self):
        config = _config.InstanceConfig()
        config.add_user('admin', sudo=True)
        self._check(config, {},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel'],
                                'sudo': 'ALL=(ALL) NOPASSWD:ALL'}]})

    def test_add_user_passwd(self):
        config = _config.InstanceConfig()
        config.add_user('admin', password_hash='123')
        self._check(config, {},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel'],
                                'passwd': '123'}]})

    def test_add_user_with_keys(self):
        config = _config.InstanceConfig(ssh_keys=['abc', 'def'])
        config.add_user('admin')
        self._check(config, {'public_keys': {'0': 'abc', '1': 'def'}},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel'],
                                'ssh_authorized_keys': ['abc', 'def']}]})



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import json
import sys

from metalsmith import _utils


def _print(msg, **kwargs):
    print(msg % kwargs)


class NullFormat(object):
    """Formatting outputting nothing.

    Used implicitly with --quiet.
    """

    def deploy(self, instance):
        pass

    def undeploy(self, node):
        pass


class DefaultFormat(object):
    """Human-readable formatter."""

    def deploy(self, instance):
        """Output result of the deploy."""
        self.show([instance])

    def undeploy(self, node):
        """Output result of undeploy."""
        if node.provision_state == 'available':
            message = "Successfully unprovisioned node %(node)s"
        else:
            message = "Unprovisioning started for node %(node)s"

        _print(message, node=_utils.log_res(node))

    def show(self, instances):
        for instance in instances:
            _print("Node %(node)s, current state is %(state)s",
                   node=_utils.log_res(instance.node),
                   state=instance.state.name)

            if instance.hostname:
                _print('* Hostname: %(hostname)s', hostname=instance.hostname)

            if instance.is_deployed:
                ips = instance.ip_addresses()
                if ips:
                    ips = '; '.join('%s=%s' % (net, ','.join(ips))
                                    for net, ips in ips.items())
                    _print('* IP addresses: %(ips)s', ips=ips)


class JsonFormat(object):
    """JSON formatter."""

    def deploy(self, instance):
        """Output result of the deploy."""
        json.dump(instance.to_dict(), sys.stdout)

    def undeploy(self, node):
        """Output result of undeploy."""
        result = {
            'node': node.to_dict()
        }
        json.dump(result, sys.stdout)

    def show(self, instances):
        """Output instance statuses."""
        json.dump({instance.hostname: instance.to_dict()
                   for instance in instances}, sys.stdout)


FORMATS = {
    'default': DefaultFormat(),
    'json': JsonFormat()
}
"""Available formatters."""


DEFAULT_FORMAT = 'default'
"""Default formatter."""


NULL_FORMAT = NullFormat()
"""Formatter outputting nothing."""



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

from metalsmith import sources


class TestDetect(testtools.TestCase):

    def test_glance(self):
        source = sources.detect('foobar')
        self.assertIsInstance(source, sources.GlanceImage)
        self.assertEqual(source.image, 'foobar')

    def test_glance_invalid_arguments(self):
        for kwargs in [{'kernel': 'foo'},
                       {'ramdisk': 'foo'},
                       {'checksum': 'foo'}]:
            self.assertRaisesRegex(ValueError, 'cannot be provided',
                                   sources.detect, 'foobar', **kwargs)

    def test_checksum_required(self):
        for tp in ('file', 'http', 'https'):
            self.assertRaisesRegex(ValueError, 'checksum is required',
                                   sources.detect, '%s://foo' % tp)

    def test_file_whole_disk(self):
        source = sources.detect('file:///image', checksum='abcd')
        self.assertIs(source.__class__, sources.FileWholeDiskImage)
        self.assertEqual(source.location, 'file:///image')
        self.assertEqual(source.checksum, 'abcd')

    def test_file_partition_disk(self):
        source = sources.detect('file:///image', checksum='abcd',
                                kernel='file:///kernel',
                                ramdisk='file:///ramdisk')
        self.assertIs(source.__class__, sources.FilePartitionImage)
        self.assertEqual(source.location, 'file:///image')
        self.assertEqual(source.checksum, 'abcd')
        self.assertEqual(source.kernel_location, 'file:///kernel')
        self.assertEqual(source.ramdisk_location, 'file:///ramdisk')

    def test_file_partition_inconsistency(self):
        for kwargs in [{'kernel': 'foo'},
                       {'ramdisk': 'foo'},
                       {'kernel': 'http://foo'},
                       {'ramdisk': 'http://foo'},
                       {'checksum': 'http://foo'}]:
            kwargs.setdefault('checksum', 'abcd')
            self.assertRaisesRegex(ValueError, 'can only be files',
                                   sources.detect, 'file:///image', **kwargs)

    def test_http_whole_disk(self):
        source = sources.detect('http:///image', checksum='abcd')
        self.assertIs(source.__class__, sources.HttpWholeDiskImage)
        self.assertEqual(source.url, 'http:///image')
        self.assertEqual(source.checksum, 'abcd')

    def test_https_whole_disk(self):
        source = sources.detect('https:///image', checksum='abcd')
        self.assertIs(source.__class__, sources.HttpWholeDiskImage)
        self.assertEqual(source.url, 'https:///image')
        self.assertEqual(source.checksum, 'abcd')

    def test_https_whole_disk_checksum(self):
        source = sources.detect('https:///image',
                                checksum='https://checksum')
        self.assertIs(source.__class__, sources.HttpWholeDiskImage)
        self.assertEqual(source.url, 'https:///image')
        self.assertEqual(source.checksum_url, 'https://checksum')

    def test_http_partition_disk(self):
        source = sources.detect('http:///image', checksum='abcd',
                                kernel='http:///kernel',
                                ramdisk='http:///ramdisk')
        self.assertIs(source.__class__, sources.HttpPartitionImage)
        self.assertEqual(source.url, 'http:///image')
        self.assertEqual(source.checksum, 'abcd')
        self.assertEqual(source.kernel_url, 'http:///kernel')
        self.assertEqual(source.ramdisk_url, 'http:///ramdisk')

    def test_https_partition_disk(self):
        source = sources.detect('https:///image', checksum='abcd',
                                # Can mix HTTP and HTTPs
                                kernel='http:///kernel',
                                ramdisk='https:///ramdisk')
        self.assertIs(source.__class__, sources.HttpPartitionImage)
        self.assertEqual(source.url, 'https:///image')
        self.assertEqual(source.checksum, 'abcd')
        self.assertEqual(source.kernel_url, 'http:///kernel')
        self.assertEqual(source.ramdisk_url, 'https:///ramdisk')

    def test_https_partition_disk_checksum(self):
        source = sources.detect('https:///image',
                                # Can mix HTTP and HTTPs
                                checksum='http://checksum',
                                kernel='http:///kernel',
                                ramdisk='https:///ramdisk')
        self.assertIs(source.__class__, sources.HttpPartitionImage)
        self.assertEqual(source.url, 'https:///image')
        self.assertEqual(source.checksum_url, 'http://checksum')
        self.assertEqual(source.kernel_url, 'http:///kernel')
        self.assertEqual(source.ramdisk_url, 'https:///ramdisk')

    def test_http_partition_inconsistency(self):
        for kwargs in [{'kernel': 'foo'},
                       {'ramdisk': 'foo'},
                       {'kernel': 'file://foo'},
                       {'ramdisk': 'file://foo'},
                       {'checksum': 'file://foo'}]:
            kwargs.setdefault('checksum', 'abcd')
            self.assertRaisesRegex(ValueError, 'can only be HTTP',
                                   sources.detect, 'http:///image', **kwargs)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from openstack import exceptions as sdk_exc
import testtools

from metalsmith import _scheduler
from metalsmith import exceptions


class TestScheduleNode(testtools.TestCase):

    def setUp(self):
        super(TestScheduleNode, self).setUp()
        self.nodes = [mock.Mock(spec=['id', 'name']) for _ in range(2)]
        self.reserver = self._reserver(lambda x: x)

    def _reserver(self, side_effect):
        reserver = mock.Mock(spec=_scheduler.Reserver)
        reserver.side_effect = side_effect
        if isinstance(side_effect, Exception):
            reserver.fail.side_effect = exceptions.ReservationFailed('fail')
        else:
            reserver.fail.side_effect = AssertionError('called fail')
        return reserver

    def _filter(self, side_effect, fail=AssertionError('called fail')):
        fltr = mock.Mock(spec=_scheduler.Filter)
        fltr.side_effect = side_effect
        fltr.fail.side_effect = fail
        return fltr

    def test_no_filters(self):
        result = _scheduler.schedule_node(self.nodes, [], self.reserver)
        self.assertIs(result, self.nodes[0])
        self.reserver.assert_called_once_with(self.nodes[0])
        self.assertFalse(self.reserver.fail.called)

    def test_dry_run(self):
        result = _scheduler.schedule_node(self.nodes, [], self.reserver,
                                          dry_run=True)
        self.assertIs(result, self.nodes[0])
        self.assertFalse(self.reserver.called)
        self.assertFalse(self.reserver.fail.called)

    def test_reservation_one_failed(self):
        reserver = self._reserver([sdk_exc.SDKException("boom"),
                                   self.nodes[1]])
        result = _scheduler.schedule_node(self.nodes, [], reserver)
        self.assertIs(result, self.nodes[1])
        self.assertEqual([mock.call(n) for n in self.nodes],
                         reserver.call_args_list)

    def test_reservation_all_failed(self):
        reserver = self._reserver(sdk_exc.SDKException("boom"))
        self.assertRaisesRegex(exceptions.ReservationFailed, 'fail',
                               _scheduler.schedule_node,
                               self.nodes, [], reserver)
        self.assertEqual([mock.call(n) for n in self.nodes],
                         reserver.call_args_list)

    def test_all_filters_pass(self):
        filters = [self._filter([True, True]) for _ in range(3)]
        result = _scheduler.schedule_node(self.nodes, filters, self.reserver)
        self.assertIs(result, self.nodes[0])
        self.reserver.assert_called_once_with(self.nodes[0])
        for fltr in filters:
            self.assertEqual([mock.call(n) for n in self.nodes],
                             fltr.call_args_list)
            self.assertFalse(fltr.fail.called)

    def test_one_node_filtered(self):
        filters = [self._filter([True, True]),
                   self._filter([False, True]),
                   self._filter([True])]
        result = _scheduler.schedule_node(self.nodes, filters, self.reserver)
        self.assertIs(result, self.nodes[1])
        self.reserver.assert_called_once_with(self.nodes[1])
        for fltr in filters:
            self.assertFalse(fltr.fail.called)
        for fltr in filters[:2]:
            self.assertEqual([mock.call(n) for n in self.nodes],
                             fltr.call_args_list)
        filters[2].assert_called_once_with(self.nodes[1])

    def test_all_nodes_filtered(self):
        filters = [self._filter([True, True]),
                   self._filter([False, True]),
                   self._filter([False], fail=RuntimeError('failed'))]
        self.assertRaisesRegex(RuntimeError, 'failed',
                               _scheduler.schedule_node,
                               self.nodes, filters, self.reserver)
        self.assertFalse(self.reserver.called)
        for fltr in filters[:2]:
            self.assertEqual([mock.call(n) for n in self.nodes],
                             fltr.call_args_list)
            self.assertFalse(fltr.fail.called)
        filters[2].assert_called_once_with(self.nodes[1])
        filters[2].fail.assert_called_once_with()


class TestCapabilitiesFilter(testtools.TestCase):

    def test_fail_no_capabilities(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute'})
        self.assertRaisesRegex(exceptions.CapabilitiesNotFound,
                               'No available nodes found with capabilities '
                               'profile=compute, existing capabilities: none',
                               fltr.fail)

    def test_nothing_requested_nothing_found(self):
        fltr = _scheduler.CapabilitiesFilter({})
        node = mock.Mock(properties={}, spec=['properties', 'name', 'id'])
        self.assertTrue(fltr(node))

    def test_matching_node(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute',
                                              'foo': 'bar'})
        node = mock.Mock(
            properties={'capabilities': 'foo:bar,profile:compute,answer:42'},
            spec=['properties', 'name', 'id'])
        self.assertTrue(fltr(node))

    def test_not_matching_node(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute',
                                              'foo': 'bar'})
        node = mock.Mock(
            properties={'capabilities': 'foo:bar,answer:42'},
            spec=['properties', 'name', 'id'])
        self.assertFalse(fltr(node))

    def test_fail_message(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute'})
        node = mock.Mock(
            properties={'capabilities': 'profile:control'},
            spec=['properties', 'name', 'id'])
        self.assertFalse(fltr(node))
        self.assertRaisesRegex(exceptions.CapabilitiesNotFound,
                               'No available nodes found with capabilities '
                               'profile=compute, existing capabilities: '
                               r'profile=control \(1 node\(s\)\)',
                               fltr.fail)

    def test_malformed_capabilities(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute'})
        for cap in ['foo,profile:control', 42, 'a:b:c']:
            node = mock.Mock(properties={'capabilities': cap},
                             spec=['properties', 'name', 'id'])
            self.assertFalse(fltr(node))
        self.assertRaisesRegex(exceptions.CapabilitiesNotFound,
                               'No available nodes found with capabilities '
                               'profile=compute, existing capabilities: none',
                               fltr.fail)


class TestTraitsFilter(testtools.TestCase):

    def test_fail_no_traits(self):
        fltr = _scheduler.TraitsFilter(['tr1', 'tr2'])
        self.assertRaisesRegex(exceptions.TraitsNotFound,
                               'No available nodes found with traits '
                               'tr1, tr2, existing traits: none',
                               fltr.fail)

    def test_no_traits(self):
        fltr = _scheduler.TraitsFilter([])
        node = mock.Mock(spec=['name', 'id'])
        self.assertTrue(fltr(node))

    def test_ok(self):
        fltr = _scheduler.TraitsFilter(['tr1', 'tr2'])
        node = mock.Mock(spec=['name', 'id', 'traits'],
                         traits=['tr3', 'tr2', 'tr1'])
        self.assertTrue(fltr(node))

    def test_missing_one(self):
        fltr = _scheduler.TraitsFilter(['tr1', 'tr2'])
        node = mock.Mock(spec=['name', 'id', 'traits'],
                         traits=['tr3', 'tr1'])
        self.assertFalse(fltr(node))

    def test_missing_all(self):
        fltr = _scheduler.TraitsFilter(['tr1', 'tr2'])
        node = mock.Mock(spec=['name', 'id', 'traits'], traits=None)
        self.assertFalse(fltr(node))


class TestIronicReserver(testtools.TestCase):

    def setUp(self):
        super(TestIronicReserver, self).setUp()
        self.node = mock.Mock(spec=['id', 'name', 'instance_info'],
                              instance_info={})
        self.api = mock.Mock(spec=['baremetal'])
        self.api.baremetal = mock.Mock(spec=['update_node', 'validate_node'])
        self.api.baremetal.update_node.side_effect = (
            lambda node, **kw: node)
        self.reserver = _scheduler.IronicReserver(self.api)

    def test_fail(self):
        self.assertRaisesRegex(exceptions.NoNodesReserved,
                               'All the candidate nodes are already reserved',
                               self.reserver.fail)

    def test_ok(self):
        self.assertEqual(self.node, self.reserver(self.node))
        self.api.baremetal.validate_node.assert_called_with(
            self.node, required=('power', 'management'))
        self.api.baremetal.update_node.assert_called_once_with(
            self.node, instance_id=self.node.id, instance_info={})

    def test_with_instance_info(self):
        self.reserver = _scheduler.IronicReserver(self.api,
                                                  {'cat': 'meow'})
        self.assertEqual(self.node, self.reserver(self.node))
        self.api.baremetal.validate_node.assert_called_with(
            self.node, required=('power', 'management'))
        self.api.baremetal.update_node.assert_called_once_with(
            self.node, instance_id=self.node.id,
            instance_info={'cat': 'meow'})

    def test_reservation_failed(self):
        self.api.baremetal.update_node.side_effect = (
            sdk_exc.SDKException('conflict'))
        self.assertRaisesRegex(sdk_exc.SDKException, 'conflict',
                               self.reserver, self.node)
        self.api.baremetal.validate_node.assert_called_with(
            self.node, required=('power', 'management'))
        self.api.baremetal.update_node.assert_called_once_with(
            self.node, instance_id=self.node.id, instance_info={})

    def test_validation_failed(self):
        self.api.baremetal.validate_node.side_effect = (
            sdk_exc.SDKException('fail'))
        self.assertRaisesRegex(exceptions.ValidationFailed, 'fail',
                               self.reserver, self.node)
        self.api.baremetal.validate_node.assert_called_with(
            self.node, required=('power', 'management'))
        self.assertFalse(self.api.baremetal.update_node.called)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import warnings

from metalsmith import _utils


_PROGRESS_STATES = frozenset(['deploying', 'wait call-back',
                              'deploy complete'])
_ACTIVE_STATES = frozenset(['active'])
_ERROR_STATES = frozenset(['error', 'deploy failed'])
_RESERVED_STATES = frozenset(['available'])


class InstanceState(enum.Enum):
    """A state of an instance."""

    DEPLOYING = 'deploying'
    """Provisioning is in progress.

    This includes the case when a node is still in the ``available`` state, but
    already has an instance associated with it.
    """

    ACTIVE = 'active'
    """The instance is provisioned."""

    MAINTENANCE = 'maintenance'
    """The instance is provisioned but is in the maintenance mode."""

    ERROR = 'error'
    """The instance has a failure."""

    UNKNOWN = 'unknown'
    """The node is in an unexpected state.

    It can be unprovisioned or modified by a third party.
    """

    @property
    def is_deployed(self):
        """Whether the state designates a finished deployment."""
        return self in _DEPLOYED_STATES

    @property
    def is_healthy(self):
        """Whether the state is considered healthy."""
        return self in _HEALTHY_STATES

    # TODO(dtantsur): remove before 1.0

    def __eq__(self, other):
        if isinstance(other, str):
            warnings.warn("Comparing states with strings is deprecated, "
                          "use InstanceState instead", DeprecationWarning)
            return self.value == other
        else:
            return super(InstanceState, self).__eq__(other)

    def __ne__(self, other):
        eq = self.__eq__(other)
        return NotImplemented if eq is NotImplemented else not eq

    def __hash__(self):
        return hash(self.value)


_HEALTHY_STATES = frozenset([InstanceState.ACTIVE, InstanceState.DEPLOYING])
_DEPLOYED_STATES = frozenset([InstanceState.ACTIVE, InstanceState.MAINTENANCE])


class Instance(object):
    """Instance status in metalsmith."""

    def __init__(self, connection, node):
        self._connection = connection
        self._uuid = node.id
        self._node = node

    @property
    def hostname(self):
        """Node's hostname."""
        return self._node.instance_info.get(_utils.GetNodeMixin.HOSTNAME_FIELD)

    def ip_addresses(self):
        """Returns IP addresses for this instance.

        :return: dict mapping network name or ID to a list of IP addresses.
        """
        result = {}
        for nic in self.nics():
            net = getattr(nic.network, 'name', None) or nic.network.id
            result.setdefault(net, []).extend(
                ip['ip_address'] for ip in nic.fixed_ips
                if ip.get('ip_address')
            )
        return result

    @property
    def is_deployed(self):
        """Whether the node is deployed."""
        return self.state.is_deployed

    @property
    def is_healthy(self):
        """Whether the instance is not at fault or maintenance."""
        return self.state.is_healthy and not self._node.is_maintenance

    def nics(self):
        """List NICs for this instance.

        :return: List of `Port` objects with additional ``network`` fields
            with full representations of their networks.
        """
        result = []
        vifs = self._connection.baremetal.list_node_vifs(self.node)
        for vif in vifs:
            port = self._connection.network.get_port(vif)
            port.network = self._connection.network.get_network(
                port.network_id)
            result.append(port)
        return result

    @property
    def node(self):
        """Underlying `Node` object."""
        return self._node

    @property
    def state(self):
        """Instance state, one of :py:class:`InstanceState`."""
        prov_state = self._node.provision_state
        if prov_state in _PROGRESS_STATES:
            return InstanceState.DEPLOYING
        # NOTE(dtantsur): include available since there is a period of time
        # between claiming the instance and starting the actual provisioning.
        elif prov_state in _RESERVED_STATES and self._node.instance_id:
            return InstanceState.DEPLOYING
        elif prov_state in _ERROR_STATES:
            return InstanceState.ERROR
        elif prov_state in _ACTIVE_STATES:
            if self._node.is_maintenance:
                return InstanceState.MAINTENANCE
            else:
                return InstanceState.ACTIVE
        else:
            return InstanceState.UNKNOWN

    def to_dict(self):
        """Convert instance to a dict."""
        return {
            'hostname': self.hostname,
            'ip_addresses': self.ip_addresses(),
            'node': self._node.to_dict(),
            'state': self.state.value,
            'uuid': self._uuid,
        }

    @property
    def uuid(self):
        """Instance UUID (the same as `Node` UUID for metalsmith)."""
        return self._uuid



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Image sources to use when provisioning nodes."""

import abc
import logging
import os

import openstack.exceptions
import requests
import six
from six.moves.urllib import parse as urlparse

from metalsmith import _utils
from metalsmith import exceptions


LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class _Source(object):

    def _validate(self, connection):
        """Validate the source."""

    @abc.abstractmethod
    def _node_updates(self, connection):
        """Updates required for a node to use this source."""


class GlanceImage(_Source):
    """Image from the OpenStack Image service."""

    def __init__(self, image):
        """Create a Glance source.

        :param image: `Image` object, ID or name.
        """
        self.image = image
        self._image_obj = None

    def _validate(self, connection):
        if self._image_obj is not None:
            return
        try:
            self._image_obj = connection.image.find_image(self.image,
                                                          ignore_missing=False)
        except openstack.exceptions.SDKException as exc:
            raise exceptions.InvalidImage(
                'Cannot find image %(image)s: %(error)s' %
                {'image': self.image, 'error': exc})

    def _node_updates(self, connection):
        self._validate(connection)
        LOG.debug('Image: %s', self._image_obj)

        updates = {
            'image_source': self._image_obj.id
        }
        for prop in ('kernel', 'ramdisk'):
            value = getattr(self._image_obj, '%s_id' % prop, None)
            if value:
                updates[prop] = value

        return updates


class HttpWholeDiskImage(_Source):
    """A whole-disk image from HTTP(s) location.

    Some deployment methods require a checksum of the image. It has to be
    provided via ``checksum`` or ``checksum_url``.

    Only ``checksum_url`` (if provided) has to be accessible from the current
    machine. Other URLs have to be accessible by the Bare Metal service (more
    specifically, by **ironic-conductor** processes).
    """

    def __init__(self, url, checksum=None, checksum_url=None):
        """Create an HTTP source.

        :param url: URL of the image.
        :param checksum: MD5 checksum of the image. Mutually exclusive with
            ``checksum_url``.
        :param checksum_url: URL of the checksum file for the image. Has to
            be in the standard format of the ``md5sum`` tool. Mutually
            exclusive with ``checksum``.
        """
        if (checksum and checksum_url) or (not checksum and not checksum_url):
            raise TypeError('Exactly one of checksum and checksum_url has '
                            'to be specified')

        self.url = url
        self.checksum = checksum
        self.checksum_url = checksum_url

    def _validate(self, connection):
        # TODO(dtantsur): should we validate image URLs here? Ironic will do it
        # as well, and images do not have to be accessible from where
        # metalsmith is running.
        if self.checksum:
            return

        try:
            response = requests.get(self.checksum_url)
            response.raise_for_status()
            checksums = response.text
        except requests.RequestException as exc:
            raise exceptions.InvalidImage(
                'Cannot download checksum file %(url)s: %(err)s' %
                {'url': self.checksum_url, 'err': exc})

        try:
            checksums = _utils.parse_checksums(checksums)
        except (ValueError, TypeError) as exc:
            raise exceptions.InvalidImage(
                'Invalid checksum file %(url)s: %(err)s' %
                {'url': self.checksum_url, 'err': exc})

        fname = os.path.basename(urlparse.urlparse(self.url).path)
        try:
            self.checksum = checksums[fname]
        except KeyError:
            raise exceptions.InvalidImage(
                'There is no image checksum for %(fname)s in %(url)s' %
                {'fname': fname, 'url': self.checksum_url})

    def _node_updates(self, connection):
        self._validate(connection)
        LOG.debug('Image: %(image)s, checksum %(checksum)s',
                  {'image': self.url, 'checksum': self.checksum})
        return {
            'image_source': self.url,
            'image_checksum': self.checksum,
        }


class HttpPartitionImage(HttpWholeDiskImage):
    """A partition image from an HTTP(s) location."""

    def __init__(self, url, kernel_url, ramdisk_url, checksum=None,
                 checksum_url=None):
        """Create an HTTP source.

        :param url: URL of the root disk image.
        :param kernel_url: URL of the kernel image.
        :param ramdisk_url: URL of the initramfs image.
        :param checksum: MD5 checksum of the root disk image. Mutually
            exclusive with ``checksum_url``.
        :param checksum_url: URL of the checksum file for the root disk image.
            Has to be in the standard format of the ``md5sum`` tool. Mutually
            exclusive with ``checksum``.
        """
        super(HttpPartitionImage, self).__init__(url, checksum=checksum,
                                                 checksum_url=checksum_url)
        self.kernel_url = kernel_url
        self.ramdisk_url = ramdisk_url

    def _node_updates(self, connection):
        updates = super(HttpPartitionImage, self)._node_updates(connection)
        updates['kernel'] = self.kernel_url
        updates['ramdisk'] = self.ramdisk_url
        return updates


class FileWholeDiskImage(_Source):
    """A whole-disk image from a local file location.

    .. warning::
        The location must be local to the **ironic-conductor** process handling
        the node, not to metalsmith itself! Since there is no easy way to
        determine which conductor handles a node, the same file must be
        available at the same location to all conductors in the same group.
    """

    def __init__(self, location, checksum):
        """Create a local file source.

        :param location: Location of the image, optionally starting with
            ``file://``.
        :param checksum: MD5 checksum of the image.
        """
        if not location.startswith('file://'):
            location = 'file://' + location
        self.location = location
        self.checksum = checksum

    def _node_updates(self, connection):
        LOG.debug('Image: %(image)s, checksum %(checksum)s',
                  {'image': self.location, 'checksum': self.checksum})
        return {
            'image_source': self.location,
            'image_checksum': self.checksum,
        }


class FilePartitionImage(FileWholeDiskImage):
    """A partition image from a local file location.

    .. warning::
        The location must be local to the **ironic-conductor** process handling
        the node, not to metalsmith itself! Since there is no easy way to
        determine which conductor handles a node, the same file must be
        available at the same location to all conductors in the same group.
    """

    def __init__(self, location, kernel_location, ramdisk_location, checksum):
        """Create a local file source.

        :param location: Location of the image, optionally starting with
            ``file://``.
        :param kernel_location: Location of the kernel of the image,
            optionally starting with ``file://``.
        :param ramdisk_location: Location of the ramdisk of the image,
            optionally starting with ``file://``.
        :param checksum: MD5 checksum of the image.
        """
        super(FilePartitionImage, self).__init__(location, checksum)
        if not kernel_location.startswith('file://'):
            kernel_location = 'file://' + kernel_location
        if not ramdisk_location.startswith('file://'):
            ramdisk_location = 'file://' + ramdisk_location
        self.kernel_location = kernel_location
        self.ramdisk_location = ramdisk_location

    def _node_updates(self, connection):
        updates = super(FilePartitionImage, self)._node_updates(connection)
        updates['kernel'] = self.kernel_location
        updates['ramdisk'] = self.ramdisk_location
        return updates


def detect(image, kernel=None, ramdisk=None, checksum=None):
    """Try detecting the correct source type from the provided information.

    .. note::
        Images without a schema are assumed to be Glance images.

    :param image: Location of the image: ``file://``, ``http://``, ``https://``
        link or a Glance image name or UUID.
    :param kernel: Location of the kernel (if present): ``file://``,
        ``http://``, ``https://`` link or a Glance image name or UUID.
    :param ramdisk: Location of the ramdisk (if present): ``file://``,
        ``http://``, ``https://`` link or a Glance image name or UUID.
    :param checksum: MD5 checksum of the image: ``http://`` or ``https://``
        link or a string.
    :return: A valid source object.
    :raises: ValueError if the given parameters do not correspond to any
        valid source.
    """
    image_type = _link_type(image)
    checksum_type = _link_type(checksum)

    if image_type == 'glance':
        if kernel or ramdisk or checksum:
            raise ValueError('kernel, image and checksum cannot be provided '
                             'for Glance images')
        else:
            return GlanceImage(image)

    kernel_type = _link_type(kernel)
    ramdisk_type = _link_type(ramdisk)
    if not checksum:
        raise ValueError('checksum is required for HTTP and file images')

    if image_type == 'file':
        if (kernel_type not in (None, 'file')
                or ramdisk_type not in (None, 'file')
                or checksum_type == 'http'):
            raise ValueError('kernal, ramdisk and checksum can only be files '
                             'for file images')

        if kernel or ramdisk:
            return FilePartitionImage(image,
                                      kernel_location=kernel,
                                      ramdisk_location=ramdisk,
                                      checksum=checksum)
        else:
            return FileWholeDiskImage(image, checksum=checksum)
    else:
        if (kernel_type not in (None, 'http')
                or ramdisk_type not in (None, 'http')
                or checksum_type == 'file'):
            raise ValueError('kernal, ramdisk and checksum can only be HTTP '
                             'links for HTTP images')

        if checksum_type == 'http':
            kwargs = {'checksum_url': checksum}
        else:
            kwargs = {'checksum': checksum}

        if kernel or ramdisk:
            return HttpPartitionImage(image,
                                      kernel_url=kernel,
                                      ramdisk_url=ramdisk,
                                      **kwargs)
        else:
            return HttpWholeDiskImage(image, **kwargs)


def _link_type(link):
    if link is None:
        return None
    elif link.startswith('http://') or link.startswith('https://'):
        return 'http'
    elif link.startswith('file://'):
        return 'file'
    else:
        return 'glance'



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Error(Exception):
    """Base class for Metalsmith errors."""


class ReservationFailed(Error):
    """Failed to reserve a suitable node.

    This is the base class for all reservation failures.
    """


class NodesNotFound(ReservationFailed):
    """Initial nodes lookup returned an empty list.

    :ivar requested_resource_class: Requested resource class.
    :ivar requested_conductor_group: Requested conductor group to pick nodes
        from.
    """

    def __init__(self, resource_class, conductor_group):
        message = "No available nodes%(rc)s found%(cg)s" % {
            'rc': ' with resource class %s' % resource_class
            if resource_class else '',
            'cg': ' in conductor group %s' % (conductor_group or '<default>')
            if conductor_group is not None else ''
        }
        self.requested_resource_class = resource_class
        self.requested_conductor_group = conductor_group
        super(NodesNotFound, self).__init__(message)


class CustomPredicateFailed(ReservationFailed):
    """Custom predicate yielded no nodes.

    :ivar nodes: List of nodes that were checked.
    """

    def __init__(self, message, nodes):
        self.nodes = nodes
        super(CustomPredicateFailed, self).__init__(message)


class CapabilitiesNotFound(ReservationFailed):
    """Requested capabilities do not match any nodes.

    :ivar requested_capabilities: Requested node's capabilities.
    """

    def __init__(self, message, capabilities):
        self.requested_capabilities = capabilities
        super(CapabilitiesNotFound, self).__init__(message)


class TraitsNotFound(ReservationFailed):
    """Requested traits do not match any nodes.

    :ivar requested_traits: Requested node's traits.
    """

    def __init__(self, message, traits):
        self.requested_traits = traits
        super(TraitsNotFound, self).__init__(message)


class ValidationFailed(ReservationFailed):
    """Validation failed for all requested nodes."""


class NoNodesReserved(ReservationFailed):
    """All nodes are already reserved or failed validation.

    :ivar nodes: List of nodes that were checked.
    """

    def __init__(self, nodes):
        self.nodes = nodes
        message = ('All the candidate nodes are already reserved '
                   'or failed validation')
        super(NoNodesReserved, self).__init__(message)


class InvalidImage(Error):
    """Requested image is invalid and cannot be used."""


class InvalidNIC(Error):
    """Requested NIC is invalid and cannot be used."""


class UnknownRootDiskSize(Error):
    """Cannot determine the root disk size."""


class InvalidNode(Error):
    """This node cannot be deployed onto."""


class DeploymentFailure(Error):
    """One or more nodes have failed the deployment.

    :ivar nodes: List of failed nodes.
    """

    def __init__(self, message, nodes):
        self.nodes = nodes
        super(DeploymentFailure, self).__init__(message)


class InvalidInstance(Error):
    """The node(s) does not have a metalsmith instance associated."""



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import collections
import logging

from openstack import exceptions as sdk_exc
import six

from metalsmith import _utils
from metalsmith import exceptions


LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class Filter(object):
    """Base class for filters."""

    @abc.abstractmethod
    def __call__(self, node):
        """Validate this node.

        :param node: Node object.
        :return: True/False
        """

    @abc.abstractmethod
    def fail(self):
        """Fail scheduling because no nodes are left.

        Must raise an exception.
        """


@six.add_metaclass(abc.ABCMeta)
class Reserver(object):
    """Base class for reservers."""

    @abc.abstractmethod
    def __call__(self, node):
        """Reserve this node.

        :param node: Node object.
        :return: updated Node object if it was reserved
        :raises: any Exception to indicate that the next node should be tried
        """

    @abc.abstractmethod
    def fail(self):
        """Fail reservation because no nodes are left.

        Must raise an exception.
        """


def schedule_node(nodes, filters, reserver, dry_run=False):
    """Schedule one node.

    :param nodes: List of input nodes.
    :param filters: List of callable Filter objects to filter/validate nodes.
        They are called in passes. If a pass yields no nodes, an error is
        raised.
    :param reserver: A callable Reserver object. Must return the updated node
        or raise an exception.
    :param dry_run: If True, reserver is not actually called.
    :return: The resulting node
    """
    for f in filters:
        f_name = f.__class__.__name__
        LOG.debug('Running filter %(filter)s on %(count)d node(s)',
                  {'filter': f_name, 'count': len(nodes)})

        nodes = list(filter(f, nodes))
        if not nodes:
            LOG.debug('Filter %s yielded no nodes', f_name)
            f.fail()
            assert False, "BUG: %s.fail did not raise" % f_name

        LOG.debug('Filter %(filter)s yielded %(count)d node(s)',
                  {'filter': f_name, 'count': len(nodes)})

    if dry_run:
        LOG.debug('Dry run, not reserving any nodes')
        return nodes[0]

    for node in nodes:
        try:
            result = reserver(node)
        except sdk_exc.SDKException as exc:
            LOG.debug('Node %(node)s was not reserved (%(exc)s), moving on '
                      'to the next one',
                      {'node': _utils.log_res(node), 'exc': exc})
        else:
            LOG.info('Node %s reserved for deployment',
                     _utils.log_res(result))
            return result

    LOG.debug('No nodes could be reserved')
    reserver.fail()
    assert False, "BUG: %s.fail did not raise" % reserver.__class__.__name__


class NodeTypeFilter(Filter):
    """Filter that checks resource class and conductor group."""

    def __init__(self, resource_class=None, conductor_group=None):
        self.resource_class = resource_class
        self.conductor_group = conductor_group

    def __call__(self, node):
        if node.instance_id:
            LOG.debug('Node %s is already reserved', _utils.log_res(node))
            return False

        if node.is_maintenance:
            LOG.debug('Node %s is in maintenance', _utils.log_res(node))
            return False

        if (self.resource_class is not None
                and node.resource_class != self.resource_class):
            LOG.debug('Resource class %(real)s does not match the expected '
                      'value of %(exp)s for node %(node)s',
                      {'node': _utils.log_res(node),
                       'exp': self.resource_class,
                       'real': node.resource_class})
            return False

        if (self.conductor_group is not None
                and node.conductor_group != self.conductor_group):
            LOG.debug('Conductor group %(real)s does not match the expected '
                      'value of %(exp)s for node %(node)s',
                      {'node': _utils.log_res(node),
                       'exp': self.conductor_group,
                       'real': node.conductor_group})
            return False

        return True

    def fail(self):
        raise exceptions.NodesNotFound(self.resource_class,
                                       self.conductor_group)


class CapabilitiesFilter(Filter):
    """Filter that checks capabilities."""

    def __init__(self, capabilities):
        self._capabilities = capabilities
        self._counter = collections.Counter()

    def __call__(self, node):
        if not self._capabilities:
            return True

        try:
            caps = _utils.get_capabilities(node)
        except Exception:
            LOG.exception('Malformed capabilities on node %(node)s: %(caps)s',
                          {'node': _utils.log_res(node),
                           'caps': node.properties.get('capabilities')})
            return False

        LOG.debug('Capabilities for node %(node)s: %(caps)s',
                  {'node': _utils.log_res(node), 'caps': caps})
        for key, value in self._capabilities.items():
            try:
                node_value = caps[key]
            except KeyError:
                LOG.debug('Node %(node)s does not have capability %(cap)s',
                          {'node': _utils.log_res(node), 'cap': key})
                return False
            else:
                self._counter["%s=%s" % (key, node_value)] += 1
                if value != node_value:
                    LOG.debug('Node %(node)s has capability %(cap)s of '
                              'value "%(node_val)s" instead of "%(expected)s"',
                              {'node': _utils.log_res(node), 'cap': key,
                               'node_val': node_value, 'expected': value})
                    return False

        return True

    def fail(self):
        existing = ", ".join("%s (%d node(s))" % item
                             for item in self._counter.items())
        requested = ', '.join("%s=%s" % item
                              for item in self._capabilities.items())
        message = ("No available nodes found with capabilities %(req)s, "
                   "existing capabilities: %(exist)s" %
                   {'req': requested, 'exist': existing or 'none'})
        raise exceptions.CapabilitiesNotFound(message, self._capabilities)


class TraitsFilter(Filter):
    """Filter that checks traits."""

    def __init__(self, traits):
        self._traits = traits
        self._counter = collections.Counter()

    def __call__(self, node):
        if not self._traits:
            return True

        traits = node.traits or []
        LOG.debug('Traits for node %(node)s: %(traits)s',
                  {'node': _utils.log_res(node), 'traits': traits})
        for trait in traits:
            self._counter[trait] += 1

        missing = set(self._traits) - set(traits)
        if missing:
            LOG.debug('Node %(node)s does not have traits %(missing)s',
                      {'node': _utils.log_res(node), 'missing': missing})
            return False

        return True

    def fail(self):
        existing = ", ".join("%s (%d node(s))" % item
                             for item in self._counter.items())
        requested = ', '.join(self._traits)
        message = ("No available nodes found with traits %(req)s, "
                   "existing traits: %(exist)s" %
                   {'req': requested, 'exist': existing or 'none'})
        raise exceptions.TraitsNotFound(message, self._traits)


class CustomPredicateFilter(Filter):

    def __init__(self, predicate):
        self.predicate = predicate
        self._failed_nodes = []

    def __call__(self, node):
        if not self.predicate(node):
            self._failed_nodes.append(node)
            return False

        return True

    def fail(self):
        message = 'No nodes satisfied the custom predicate %s' % self.predicate
        raise exceptions.CustomPredicateFailed(message, self._failed_nodes)


class IronicReserver(Reserver):

    def __init__(self, connection, instance_info=None):
        self._connection = connection
        self._failed_nodes = []
        self._iinfo = instance_info or {}

    def validate(self, node):
        try:
            self._connection.baremetal.validate_node(
                node, required=('power', 'management'))
        except sdk_exc.SDKException as exc:
            message = ('Node %(node)s failed validation: %(err)s' %
                       {'node': _utils.log_res(node), 'err': exc})
            LOG.warning(message)
            raise exceptions.ValidationFailed(message)

    def __call__(self, node):
        try:
            self.validate(node)
            iinfo = dict(node.instance_info or {}, **self._iinfo)
            return self._connection.baremetal.update_node(
                node, instance_id=node.id, instance_info=iinfo)
        except sdk_exc.SDKException:
            self._failed_nodes.append(node)
            raise

    def fail(self):
        raise exceptions.NoNodesReserved(self._failed_nodes)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from metalsmith import _utils
from metalsmith import exceptions


class TestIsHostnameSafe(testtools.TestCase):

    def test_valid(self):
        self.assertTrue(_utils.is_hostname_safe('spam'))
        self.assertTrue(_utils.is_hostname_safe('spAm'))
        self.assertTrue(_utils.is_hostname_safe('SPAM'))
        self.assertTrue(_utils.is_hostname_safe('spam-eggs'))
        self.assertTrue(_utils.is_hostname_safe('spam.eggs'))
        self.assertTrue(_utils.is_hostname_safe('9spam'))
        self.assertTrue(_utils.is_hostname_safe('spam7'))
        self.assertTrue(_utils.is_hostname_safe('br34kf4st'))
        self.assertTrue(_utils.is_hostname_safe('s' * 63))
        self.assertTrue(_utils.is_hostname_safe('www.example.com'))
        long_str = 'a' * 63 + '.' + 'b' * 63 + '.' + 'c' * 63 + '.' + 'd' * 63
        self.assertTrue(_utils.is_hostname_safe(long_str))

    def test_invalid(self):
        self.assertFalse(_utils.is_hostname_safe('-spam'))
        self.assertFalse(_utils.is_hostname_safe('spam-'))
        self.assertFalse(_utils.is_hostname_safe('spam_eggs'))
        self.assertFalse(_utils.is_hostname_safe('spam eggs'))
        self.assertFalse(_utils.is_hostname_safe('$pam'))
        self.assertFalse(_utils.is_hostname_safe('egg$'))
        self.assertFalse(_utils.is_hostname_safe('spam#eggs'))
        self.assertFalse(_utils.is_hostname_safe(' eggs'))
        self.assertFalse(_utils.is_hostname_safe('spam '))
        self.assertFalse(_utils.is_hostname_safe('s' * 64))
        self.assertFalse(_utils.is_hostname_safe(''))
        self.assertFalse(_utils.is_hostname_safe(None))
        self.assertFalse(_utils.is_hostname_safe('www.nothere.com_'))
        self.assertFalse(_utils.is_hostname_safe('www.nothere_.com'))
        self.assertFalse(_utils.is_hostname_safe('www..nothere.com'))
        self.assertFalse(_utils.is_hostname_safe('www.-nothere.com'))
        long_str = 'a' * 63 + '.' + 'b' * 63 + '.' + 'c' * 63 + '.' + 'd' * 63
        self.assertFalse(_utils.is_hostname_safe(long_str + '.'))
        self.assertFalse(_utils.is_hostname_safe('a' * 255))
        # These are valid domain names, but not hostnames (RFC 1123)
        self.assertFalse(_utils.is_hostname_safe('www.example.com.'))
        self.assertFalse(_utils.is_hostname_safe('http._sctp.www.example.com'))
        self.assertFalse(_utils.is_hostname_safe('mail.pets_r_us.net'))
        self.assertFalse(_utils.is_hostname_safe('mail-server-15.my_host.org'))
        # RFC 952 forbids single-character hostnames
        self.assertFalse(_utils.is_hostname_safe('s'))

    def test_not_none(self):
        # Need to ensure a binary response for success or fail
        self.assertIsNotNone(_utils.is_hostname_safe('spam'))
        self.assertIsNotNone(_utils.is_hostname_safe('-spam'))


class TestGetNodeMixin(testtools.TestCase):
    def setUp(self):
        super(TestGetNodeMixin, self).setUp()
        self.mixin = _utils.GetNodeMixin()
        self.mixin.connection = mock.Mock(spec=['baremetal'])
        self.api = self.mixin.connection.baremetal

    def test__get_node_with_node(self):
        node = mock.Mock(spec=['id', 'name'])
        result = self.mixin._get_node(node)
        self.assertIs(result, node)
        self.assertFalse(self.api.get_node.called)

    def test__get_node_with_node_refresh(self):
        node = mock.Mock(spec=['id', 'name'])
        result = self.mixin._get_node(node, refresh=True)
        self.assertIs(result, self.api.get_node.return_value)
        self.api.get_node.assert_called_once_with(node.id)

    def test__get_node_with_instance(self):
        node = mock.Mock(spec=['uuid', 'node'])
        result = self.mixin._get_node(node)
        self.assertIs(result, node.node)
        self.assertFalse(self.api.get_node.called)

    def test__get_node_with_instance_refresh(self):
        node = mock.Mock(spec=['uuid', 'node'])
        result = self.mixin._get_node(node, refresh=True)
        self.assertIs(result, self.api.get_node.return_value)
        self.api.get_node.assert_called_once_with(node.node.id)

    def test__get_node_with_string(self):
        result = self.mixin._get_node('node')
        self.assertIs(result, self.api.get_node.return_value)
        self.api.get_node.assert_called_once_with('node')

    def test__get_node_with_string_hostname_allowed(self):
        nodes = [
            mock.Mock(instance_info={'metalsmith_hostname': host})
            for host in ['host1', 'host2', 'host3']
        ]
        self.api.nodes.return_value = nodes

        result = self.mixin._get_node('host2', accept_hostname=True)
        self.assertIs(result, self.api.get_node.return_value)
        self.api.get_node.assert_called_once_with(nodes[1].id)

    def test__get_node_with_string_hostname_allowed_fallback(self):
        nodes = [
            mock.Mock(instance_info={'metalsmith_hostname': host})
            for host in ['host1', 'host2', 'host3']
        ]
        self.api.nodes.return_value = nodes

        result = self.mixin._get_node('node', accept_hostname=True)
        self.assertIs(result, self.api.get_node.return_value)
        self.api.get_node.assert_called_once_with('node')

    def test__get_node_with_string_hostname_not_unique(self):
        nodes = [
            mock.Mock(instance_info={'metalsmith_hostname': host})
            for host in ['host1', 'host2', 'host2']
        ]
        self.api.nodes.return_value = nodes

        self.assertRaises(exceptions.Error,
                          self.mixin._get_node,
                          'host2', accept_hostname=True)
        self.assertFalse(self.api.get_node.called)



# Copyright 2015-2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import tempfile

import fixtures
import mock
import six
import testtools

from metalsmith import _cmd
from metalsmith import _config
from metalsmith import _instance
from metalsmith import _provisioner
from metalsmith import sources


@mock.patch.object(_provisioner, 'Provisioner', autospec=True)
class TestDeploy(testtools.TestCase):
    def setUp(self):
        super(TestDeploy, self).setUp()
        self.print_fixture = self.useFixture(fixtures.MockPatch(
            'metalsmith._format._print', autospec=True))
        self.mock_print = self.print_fixture.mock

        self.os_conf_fixture = self.useFixture(fixtures.MockPatchObject(
            _cmd.os_config, 'OpenStackConfig', autospec=True))
        self.mock_os_conf = self.os_conf_fixture.mock

    def _check(self, mock_pr, args, reserve_args, provision_args,
               dry_run=False):
        reserve_defaults = dict(resource_class='compute',
                                conductor_group=None,
                                capabilities={},
                                traits=[],
                                candidates=None)
        reserve_defaults.update(reserve_args)

        provision_defaults = dict(image=mock.ANY,
                                  nics=[{'network': 'mynet'}],
                                  root_size_gb=None,
                                  swap_size_mb=None,
                                  config=mock.ANY,
                                  hostname=None,
                                  netboot=False,
                                  wait=1800)
        provision_defaults.update(provision_args)

        _cmd.main(args)

        mock_pr.assert_called_once_with(
            cloud_region=self.mock_os_conf.return_value.get_one.return_value,
            dry_run=dry_run)
        mock_pr.return_value.reserve_node.assert_called_once_with(
            **reserve_defaults)
        mock_pr.return_value.provision_node.assert_called_once_with(
            mock_pr.return_value.reserve_node.return_value,
            **provision_defaults)

    @mock.patch.object(_cmd, 'logging', autospec=True)
    def test_args_ok(self, mock_log, mock_pr):
        instance = mock_pr.return_value.provision_node.return_value
        instance.create_autospec(_instance.Instance)
        instance.node.name = None
        instance.node.id = '123'
        instance.state = _instance.InstanceState.ACTIVE
        instance.is_deployed = True
        instance.ip_addresses.return_value = {'private': ['1.2.3.4']}
        instance.hostname = None

        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        config = mock_pr.return_value.provision_node.call_args[1]['config']
        self.assertEqual([], config.ssh_keys)
        mock_log.basicConfig.assert_called_once_with(level=mock_log.WARNING,
                                                     format=mock.ANY)

        source = mock_pr.return_value.provision_node.call_args[1]['image']
        self.assertIsInstance(source, sources.GlanceImage)
        self.assertEqual("myimg", source.image)
        self.assertEqual(
            mock.call('metalsmith').setLevel(mock_log.WARNING).call_list() +
            mock.call(_cmd._URLLIB3_LOGGER).setLevel(
                mock_log.CRITICAL).call_list(),
            mock_log.getLogger.mock_calls)

        self.mock_print.assert_has_calls([
            mock.call(mock.ANY, node='123', state='ACTIVE'),
            mock.call(mock.ANY, ips='private=1.2.3.4')
        ])

    @mock.patch.object(_cmd, 'logging', autospec=True)
    def test_args_json_format(self, mock_log, mock_pr):
        instance = mock_pr.return_value.provision_node.return_value
        instance.create_autospec(_instance.Instance)
        instance.to_dict.return_value = {'node': 'dict'}

        args = ['--format', 'json', 'deploy', '--network', 'mynet',
                '--image', 'myimg', '--resource-class', 'compute']
        fake_io = six.StringIO()
        with mock.patch('sys.stdout', fake_io):
            self._check(mock_pr, args, {}, {})
            self.assertEqual(json.loads(fake_io.getvalue()),
                             {'node': 'dict'})

        mock_log.basicConfig.assert_called_once_with(level=mock_log.WARNING,
                                                     format=mock.ANY)
        self.assertEqual(
            mock.call('metalsmith').setLevel(mock_log.WARNING).call_list() +
            mock.call(_cmd._URLLIB3_LOGGER).setLevel(
                mock_log.CRITICAL).call_list(),
            mock_log.getLogger.mock_calls)

    def test_no_ips(self, mock_pr):
        instance = mock_pr.return_value.provision_node.return_value
        instance.create_autospec(_instance.Instance)
        instance.is_deployed = True
        instance.ip_addresses.return_value = {}
        instance.node.name = None
        instance.node.id = '123'
        instance.state = _instance.InstanceState.ACTIVE
        instance.hostname = None

        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        self.mock_print.assert_called_once_with(mock.ANY, node='123',
                                                state='ACTIVE'),

    def test_not_deployed_no_ips(self, mock_pr):
        instance = mock_pr.return_value.provision_node.return_value
        instance.create_autospec(_instance.Instance)
        instance.is_deployed = False
        instance.node.name = None
        instance.node.id = '123'
        instance.state = _instance.InstanceState.DEPLOYING
        instance.hostname = None

        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        self.mock_print.assert_called_once_with(mock.ANY, node='123',
                                                state='DEPLOYING'),

    @mock.patch.object(_cmd.LOG, 'info', autospec=True)
    def test_no_logs_not_deployed(self, mock_log, mock_pr):
        instance = mock_pr.return_value.provision_node.return_value
        instance.create_autospec(_instance.Instance)
        instance.is_deployed = False

        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        self.assertFalse(mock_log.called)
        self.assertFalse(instance.ip_addresses.called)

    def test_args_dry_run(self, mock_pr):
        args = ['--dry-run', 'deploy', '--network', 'mynet',
                '--image', 'myimg', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {}, dry_run=True)

    @mock.patch.object(_cmd, 'logging', autospec=True)
    def test_args_debug(self, mock_log, mock_pr):
        args = ['--debug', 'deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        mock_log.basicConfig.assert_called_once_with(level=mock_log.DEBUG,
                                                     format=mock.ANY)
        self.assertEqual(
            mock.call('metalsmith').setLevel(mock_log.DEBUG).call_list() +
            mock.call(_cmd._URLLIB3_LOGGER).setLevel(
                mock_log.INFO).call_list(),
            mock_log.getLogger.mock_calls)

    @mock.patch.object(_cmd, 'logging', autospec=True)
    def test_args_quiet(self, mock_log, mock_pr):
        args = ['--quiet', 'deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        mock_log.basicConfig.assert_called_once_with(level=mock_log.CRITICAL,
                                                     format=mock.ANY)
        self.assertEqual(
            mock.call('metalsmith').setLevel(mock_log.CRITICAL).call_list() +
            mock.call(_cmd._URLLIB3_LOGGER).setLevel(
                mock_log.CRITICAL).call_list(),
            mock_log.getLogger.mock_calls)

        self.assertFalse(self.mock_print.called)

    @mock.patch.object(_cmd, 'logging', autospec=True)
    def test_args_verbose_1(self, mock_log, mock_pr):
        args = ['-v', 'deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        mock_log.basicConfig.assert_called_once_with(level=mock_log.WARNING,
                                                     format=mock.ANY)
        self.assertEqual(
            mock.call('metalsmith').setLevel(mock_log.INFO).call_list() +
            mock.call(_cmd._URLLIB3_LOGGER).setLevel(
                mock_log.CRITICAL).call_list(),
            mock_log.getLogger.mock_calls)

    @mock.patch.object(_cmd, 'logging', autospec=True)
    def test_args_verbose_2(self, mock_log, mock_pr):
        args = ['-vv', 'deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        mock_log.basicConfig.assert_called_once_with(level=mock_log.INFO,
                                                     format=mock.ANY)
        self.assertEqual(
            mock.call('metalsmith').setLevel(mock_log.DEBUG).call_list() +
            mock.call(_cmd._URLLIB3_LOGGER).setLevel(
                mock_log.CRITICAL).call_list(),
            mock_log.getLogger.mock_calls)

    @mock.patch.object(_cmd, 'logging', autospec=True)
    def test_args_verbose_3(self, mock_log, mock_pr):
        args = ['-vvv', 'deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        mock_log.basicConfig.assert_called_once_with(level=mock_log.DEBUG,
                                                     format=mock.ANY)
        self.assertEqual(
            mock.call('metalsmith').setLevel(mock_log.DEBUG).call_list() +
            mock.call(_cmd._URLLIB3_LOGGER).setLevel(
                mock_log.INFO).call_list(),
            mock_log.getLogger.mock_calls)

    @mock.patch.object(_cmd.LOG, 'critical', autospec=True)
    def test_reservation_failure(self, mock_log, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        failure = RuntimeError('boom')
        mock_pr.return_value.reserve_node.side_effect = failure
        self.assertRaises(SystemExit, _cmd.main, args)
        mock_log.assert_called_once_with('%s', failure, exc_info=False)

    @mock.patch.object(_cmd.LOG, 'critical', autospec=True)
    def test_deploy_failure(self, mock_log, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--resource-class', 'compute']
        failure = RuntimeError('boom')
        mock_pr.return_value.provision_node.side_effect = failure
        self.assertRaises(SystemExit, _cmd.main, args)
        mock_log.assert_called_once_with('%s', failure, exc_info=False)

    @mock.patch.object(_cmd.LOG, 'critical', autospec=True)
    def test_invalid_hostname(self, mock_log, mock_pr):
        args = ['deploy', '--hostname', 'n_1', '--image', 'myimg',
                '--resource-class', 'compute']
        self.assertRaises(SystemExit, _cmd.main, args)
        self.assertTrue(mock_log.called)

    def test_args_capabilities(self, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--capability', 'foo=bar', '--capability', 'answer=42',
                '--resource-class', 'compute']
        self._check(mock_pr, args,
                    {'capabilities': {'foo': 'bar', 'answer': '42'}}, {})

    def test_args_traits(self, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--trait', 'foo:bar', '--trait', 'answer:42',
                '--resource-class', 'compute']
        self._check(mock_pr, args,
                    {'traits': ['foo:bar', 'answer:42']}, {})

    def test_args_configdrive(self, mock_pr):
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(b'foo\n')
            fp.flush()

            args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                    '--ssh-public-key', fp.name, '--resource-class', 'compute']
            self._check(mock_pr, args, {}, {})

        config = mock_pr.return_value.provision_node.call_args[1]['config']
        self.assertEqual(['foo'], config.ssh_keys)

    @mock.patch.object(_config.InstanceConfig, 'add_user', autospec=True)
    def test_args_user_name(self, mock_add_user, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--user-name', 'banana', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {})

        config = mock_pr.return_value.provision_node.call_args[1]['config']
        self.assertEqual([], config.ssh_keys)
        mock_add_user.assert_called_once_with(config, 'banana', sudo=False)

    @mock.patch.object(_config.InstanceConfig, 'add_user', autospec=True)
    def test_args_user_name_with_sudo(self, mock_add_user, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--user-name', 'banana', '--resource-class', 'compute',
                '--passwordless-sudo']
        self._check(mock_pr, args, {}, {})

        config = mock_pr.return_value.provision_node.call_args[1]['config']
        self.assertEqual([], config.ssh_keys)
        mock_add_user.assert_called_once_with(config, 'banana', sudo=True)

    def test_args_port(self, mock_pr):
        args = ['deploy', '--port', 'myport', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'nics': [{'port': 'myport'}]})

    def test_args_no_nics(self, mock_pr):
        args = ['deploy', '--image', 'myimg', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'nics': None})

    def test_args_networks_and_ports(self, mock_pr):
        args = ['deploy', '--network', 'net1', '--port', 'port1',
                '--port', 'port2', '--network', 'net2',
                '--image', 'myimg', '--resource-class', 'compute']
        self._check(mock_pr, args, {},
                    {'nics': [{'network': 'net1'}, {'port': 'port1'},
                              {'port': 'port2'}, {'network': 'net2'}]})

    def test_args_ips(self, mock_pr):
        args = ['deploy', '--image', 'myimg', '--resource-class', 'compute',
                '--ip', 'private:10.0.0.2', '--ip', 'public:8.0.8.0']
        self._check(mock_pr, args, {},
                    {'nics': [{'network': 'private', 'fixed_ip': '10.0.0.2'},
                              {'network': 'public', 'fixed_ip': '8.0.8.0'}]})

    def test_args_subnet(self, mock_pr):
        args = ['deploy', '--subnet', 'mysubnet', '--image', 'myimg',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'nics': [{'subnet': 'mysubnet'}]})

    def test_args_bad_ip(self, mock_pr):
        args = ['deploy', '--image', 'myimg', '--resource-class', 'compute',
                '--ip', 'private:10.0.0.2', '--ip', 'public']
        self.assertRaises(SystemExit, _cmd.main, args)
        self.assertFalse(mock_pr.return_value.reserve_node.called)
        self.assertFalse(mock_pr.return_value.provision_node.called)

    def test_args_hostname(self, mock_pr):
        instance = mock_pr.return_value.provision_node.return_value
        instance.create_autospec(_instance.Instance)
        instance.is_deployed = True
        instance.node.name = None
        instance.node.id = '123'
        instance.state = _instance.InstanceState.ACTIVE
        instance.hostname = 'host'
        instance.ip_addresses.return_value = {'private': ['1.2.3.4']}

        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--hostname', 'host', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'hostname': 'host'})

        self.mock_print.assert_has_calls([
            mock.call(mock.ANY, node='123', state='ACTIVE'),
            mock.call(mock.ANY, hostname='host'),
            mock.call(mock.ANY, ips='private=1.2.3.4')
        ])

    def test_args_with_candidates(self, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--candidate', 'node1', '--candidate', 'node2',
                '--resource-class', 'compute']
        self._check(mock_pr, args, {'candidates': ['node1', 'node2']}, {})

    def test_args_conductor_group(self, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--conductor-group', 'loc1', '--resource-class', 'compute']
        self._check(mock_pr, args, {'conductor_group': 'loc1'}, {})

    def test_args_http_image_with_checksum(self, mock_pr):
        args = ['deploy', '--image', 'https://example.com/image.img',
                '--image-checksum', '95e750180c7921ea0d545c7165db66b8',
                '--network', 'mynet', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'image': mock.ANY})

        source = mock_pr.return_value.provision_node.call_args[1]['image']
        self.assertIsInstance(source, sources.HttpWholeDiskImage)
        self.assertEqual('https://example.com/image.img', source.url)
        self.assertEqual('95e750180c7921ea0d545c7165db66b8', source.checksum)

    def test_args_http_image_with_checksum_url(self, mock_pr):
        args = ['deploy', '--image', 'http://example.com/image.img',
                '--image-checksum', 'http://example.com/CHECKSUMS',
                '--network', 'mynet', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'image': mock.ANY})

        source = mock_pr.return_value.provision_node.call_args[1]['image']
        self.assertIsInstance(source, sources.HttpWholeDiskImage)
        self.assertEqual('http://example.com/image.img', source.url)
        self.assertEqual('http://example.com/CHECKSUMS', source.checksum_url)

    @mock.patch.object(_cmd.LOG, 'critical', autospec=True)
    def test_args_http_image_without_checksum(self, mock_log, mock_pr):
        args = ['deploy', '--image', 'http://example.com/image.img',
                '--resource-class', 'compute']
        self.assertRaises(SystemExit, _cmd.main, args)
        self.assertTrue(mock_log.called)
        self.assertFalse(mock_pr.return_value.reserve_node.called)
        self.assertFalse(mock_pr.return_value.provision_node.called)

    def test_args_http_partition_image(self, mock_pr):
        args = ['deploy', '--image', 'https://example.com/image.img',
                '--image-kernel', 'https://example.com/kernel',
                '--image-ramdisk', 'https://example.com/ramdisk',
                '--image-checksum', '95e750180c7921ea0d545c7165db66b8',
                '--network', 'mynet', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'image': mock.ANY})

        source = mock_pr.return_value.provision_node.call_args[1]['image']
        self.assertIsInstance(source, sources.HttpPartitionImage)
        self.assertEqual('https://example.com/image.img', source.url)
        self.assertEqual('https://example.com/kernel', source.kernel_url)
        self.assertEqual('https://example.com/ramdisk', source.ramdisk_url)
        self.assertEqual('95e750180c7921ea0d545c7165db66b8', source.checksum)

    def test_args_file_whole_disk_image(self, mock_pr):
        args = ['deploy', '--image', 'file:///var/lib/ironic/image.img',
                '--image-checksum', '95e750180c7921ea0d545c7165db66b8',
                '--network', 'mynet', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'image': mock.ANY})

        source = mock_pr.return_value.provision_node.call_args[1]['image']
        self.assertIsInstance(source, sources.FileWholeDiskImage)
        self.assertEqual('file:///var/lib/ironic/image.img', source.location)
        self.assertEqual('95e750180c7921ea0d545c7165db66b8', source.checksum)

    def test_args_file_partition_disk_image(self, mock_pr):
        args = ['deploy', '--image', 'file:///var/lib/ironic/image.img',
                '--image-kernel', 'file:///var/lib/ironic/image.vmlinuz',
                '--image-ramdisk', 'file:///var/lib/ironic/image.initrd',
                '--image-checksum', '95e750180c7921ea0d545c7165db66b8',
                '--network', 'mynet', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'image': mock.ANY})

        source = mock_pr.return_value.provision_node.call_args[1]['image']
        self.assertIsInstance(source, sources.FilePartitionImage)
        self.assertEqual('file:///var/lib/ironic/image.img', source.location)
        self.assertEqual('file:///var/lib/ironic/image.vmlinuz',
                         source.kernel_location)
        self.assertEqual('file:///var/lib/ironic/image.initrd',
                         source.ramdisk_location)
        self.assertEqual('95e750180c7921ea0d545c7165db66b8', source.checksum)

    @mock.patch.object(_cmd.LOG, 'critical', autospec=True)
    def test_args_file_image_without_checksum(self, mock_log, mock_pr):
        args = ['deploy', '--image', 'file:///var/lib/ironic/image.img',
                '--resource-class', 'compute']
        self.assertRaises(SystemExit, _cmd.main, args)
        self.assertTrue(mock_log.called)
        self.assertFalse(mock_pr.return_value.reserve_node.called)
        self.assertFalse(mock_pr.return_value.provision_node.called)

    @mock.patch.object(_cmd.LOG, 'critical', autospec=True)
    def test_args_file_image_with_incorrect_kernel(self, mock_log, mock_pr):
        args = ['deploy', '--image', 'file:///var/lib/ironic/image.img',
                '--image-kernel', 'http://example.com/image.vmlinuz',
                '--image-checksum', '95e750180c7921ea0d545c7165db66b8',
                '--resource-class', 'compute']
        self.assertRaises(SystemExit, _cmd.main, args)
        self.assertTrue(mock_log.called)
        self.assertFalse(mock_pr.return_value.reserve_node.called)
        self.assertFalse(mock_pr.return_value.provision_node.called)

    def test_args_custom_wait(self, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--wait', '3600', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'wait': 3600})

    def test_args_no_wait(self, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--no-wait', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'wait': None})

    def test_with_root_size(self, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--root-size', '100', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'root_size_gb': 100})

    def test_with_swap_size(self, mock_pr):
        args = ['deploy', '--network', 'mynet', '--image', 'myimg',
                '--swap-size', '4096', '--resource-class', 'compute']
        self._check(mock_pr, args, {}, {'swap_size_mb': 4096})


@mock.patch.object(_provisioner, 'Provisioner', autospec=True)
@mock.patch.object(_cmd.os_config, 'OpenStackConfig', autospec=True)
class TestUndeploy(testtools.TestCase):
    def setUp(self):
        super(TestUndeploy, self).setUp()
        self.print_fixture = self.useFixture(fixtures.MockPatch(
            'metalsmith._format._print', autospec=True))
        self.mock_print = self.print_fixture.mock

    def test_ok(self, mock_os_conf, mock_pr):
        node = mock_pr.return_value.unprovision_node.return_value
        node.id = '123'
        node.name = None
        node.provision_state = 'cleaning'

        args = ['undeploy', '123456']
        _cmd.main(args)

        mock_pr.assert_called_once_with(
            cloud_region=mock_os_conf.return_value.get_one.return_value,
            dry_run=False)
        mock_pr.return_value.unprovision_node.assert_called_once_with(
            '123456', wait=None
        )
        self.mock_print.assert_called_once_with(
            'Unprovisioning started for node %(node)s',
            node='123')

    def test_custom_wait(self, mock_os_conf, mock_pr):
        node = mock_pr.return_value.unprovision_node.return_value
        node.id = '123'
        node.name = None
        node.provision_state = 'available'

        args = ['undeploy', '--wait', '1800', '123456']
        _cmd.main(args)
        mock_pr.assert_called_once_with(
            cloud_region=mock_os_conf.return_value.get_one.return_value,
            dry_run=False)
        mock_pr.return_value.unprovision_node.assert_called_once_with(
            '123456', wait=1800
        )
        self.mock_print.assert_called_once_with(
            'Successfully unprovisioned node %(node)s',
            node='123')

    def test_dry_run(self, mock_os_conf, mock_pr):
        args = ['--dry-run', 'undeploy', '123456']
        _cmd.main(args)
        mock_pr.assert_called_once_with(
            cloud_region=mock_os_conf.return_value.get_one.return_value,
            dry_run=True)
        mock_pr.return_value.unprovision_node.assert_called_once_with(
            '123456', wait=None
        )

    def test_quiet(self, mock_os_conf, mock_pr):
        args = ['--quiet', 'undeploy', '123456']
        _cmd.main(args)

        mock_pr.assert_called_once_with(
            cloud_region=mock_os_conf.return_value.get_one.return_value,
            dry_run=False)
        mock_pr.return_value.unprovision_node.assert_called_once_with(
            '123456', wait=None
        )
        self.assertFalse(self.mock_print.called)

    def test_json(self, mock_os_conf, mock_pr):
        node = mock_pr.return_value.unprovision_node.return_value
        node.to_dict.return_value = {'node': 'dict'}

        args = ['--format', 'json', 'undeploy', '123456']
        fake_io = six.StringIO()
        with mock.patch('sys.stdout', fake_io):
            _cmd.main(args)
            self.assertEqual(json.loads(fake_io.getvalue()),
                             {'node': {'node': 'dict'}})

        mock_pr.assert_called_once_with(
            cloud_region=mock_os_conf.return_value.get_one.return_value,
            dry_run=False)
        mock_pr.return_value.unprovision_node.assert_called_once_with(
            '123456', wait=None
        )


@mock.patch.object(_provisioner, 'Provisioner', autospec=True)
@mock.patch.object(_cmd.os_config, 'OpenStackConfig', autospec=True)
class TestShowWait(testtools.TestCase):
    def setUp(self):
        super(TestShowWait, self).setUp()
        self.print_fixture = self.useFixture(fixtures.MockPatch(
            'metalsmith._format._print', autospec=True))
        self.mock_print = self.print_fixture.mock
        self.instances = [
            mock.Mock(
                spec=_instance.Instance, hostname=hostname,
                uuid=hostname[-1], is_deployed=(hostname[-1] == '1'),
                state=_instance.InstanceState.ACTIVE
                if hostname[-1] == '1' else _instance.InstanceState.DEPLOYING,
                **{'ip_addresses.return_value': {'private': ['1.2.3.4']}})
            for hostname in ['hostname1', 'hostname2']
        ]
        for inst in self.instances:
            inst.node.id = inst.uuid
            inst.node.name = 'name-%s' % inst.uuid
            inst.to_dict.return_value = {inst.node.id: inst.node.name}

    def test_show(self, mock_os_conf, mock_pr):
        mock_pr.return_value.show_instances.return_value = self.instances
        args = ['show', 'uuid1', 'hostname2']
        _cmd.main(args)

        self.mock_print.assert_has_calls([
            mock.call(mock.ANY, node='name-1 (UUID 1)', state='ACTIVE'),
            mock.call(mock.ANY, hostname='hostname1'),
            mock.call(mock.ANY, ips='private=1.2.3.4'),
            mock.call(mock.ANY, node='name-2 (UUID 2)', state='DEPLOYING'),
            mock.call(mock.ANY, hostname='hostname2'),
        ])
        mock_pr.return_value.show_instances.assert_called_once_with(
            ['uuid1', 'hostname2'])

    def test_list(self, mock_os_conf, mock_pr):
        mock_pr.return_value.list_instances.return_value = self.instances
        args = ['list']
        _cmd.main(args)

        self.mock_print.assert_has_calls([
            mock.call(mock.ANY, node='name-1 (UUID 1)', state='ACTIVE'),
            mock.call(mock.ANY, hostname='hostname1'),
            mock.call(mock.ANY, ips='private=1.2.3.4'),
            mock.call(mock.ANY, node='name-2 (UUID 2)', state='DEPLOYING'),
            mock.call(mock.ANY, hostname='hostname2'),
        ])
        mock_pr.return_value.list_instances.assert_called_once_with()

    def test_wait(self, mock_os_conf, mock_pr):
        mock_pr.return_value.wait_for_provisioning.return_value = (
            self.instances)
        args = ['wait', 'uuid1', 'hostname2']
        _cmd.main(args)

        self.mock_print.assert_has_calls([
            mock.call(mock.ANY, node='name-1 (UUID 1)', state='ACTIVE'),
            mock.call(mock.ANY, hostname='hostname1'),
            mock.call(mock.ANY, ips='private=1.2.3.4'),
            mock.call(mock.ANY, node='name-2 (UUID 2)', state='DEPLOYING'),
            mock.call(mock.ANY, hostname='hostname2'),
        ])
        mock_pr.return_value.wait_for_provisioning.assert_called_once_with(
            ['uuid1', 'hostname2'], timeout=None)

    def test_wait_custom_timeout(self, mock_os_conf, mock_pr):
        mock_pr.return_value.wait_for_provisioning.return_value = (
            self.instances)
        args = ['wait', '--timeout', '42', 'uuid1', 'hostname2']
        _cmd.main(args)

        self.mock_print.assert_has_calls([
            mock.call(mock.ANY, node='name-1 (UUID 1)', state='ACTIVE'),
            mock.call(mock.ANY, hostname='hostname1'),
            mock.call(mock.ANY, ips='private=1.2.3.4'),
            mock.call(mock.ANY, node='name-2 (UUID 2)', state='DEPLOYING'),
            mock.call(mock.ANY, hostname='hostname2'),
        ])
        mock_pr.return_value.wait_for_provisioning.assert_called_once_with(
            ['uuid1', 'hostname2'], timeout=42)

    def test_show_json(self, mock_os_conf, mock_pr):
        mock_pr.return_value.show_instances.return_value = self.instances
        args = ['--format', 'json', 'show', 'uuid1', 'hostname2']

        fake_io = six.StringIO()
        with mock.patch('sys.stdout', fake_io):
            _cmd.main(args)
            self.assertEqual(json.loads(fake_io.getvalue()),
                             {'hostname1': {'1': 'name-1'},
                              'hostname2': {'2': 'name-2'}})

    def test_list_json(self, mock_os_conf, mock_pr):
        mock_pr.return_value.list_instances.return_value = self.instances
        args = ['--format', 'json', 'list']

        fake_io = six.StringIO()
        with mock.patch('sys.stdout', fake_io):
            _cmd.main(args)
            self.assertEqual(json.loads(fake_io.getvalue()),
                             {'hostname1': {'1': 'name-1'},
                              'hostname2': {'2': 'name-2'}})
        mock_pr.return_value.list_instances.assert_called_once_with()

    def test_wait_json(self, mock_os_conf, mock_pr):
        mock_pr.return_value.wait_for_provisioning.return_value = (
            self.instances)
        args = ['--format', 'json', 'wait', 'uuid1', 'hostname2']

        fake_io = six.StringIO()
        with mock.patch('sys.stdout', fake_io):
            _cmd.main(args)
            self.assertEqual(json.loads(fake_io.getvalue()),
                             {'hostname1': {'1': 'name-1'},
                              'hostname2': {'2': 'name-2'}})



# Copyright 2015-2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import sys

from openstack import config as os_config

from metalsmith import _config
from metalsmith import _format
from metalsmith import _provisioner
from metalsmith import _utils
from metalsmith import sources


LOG = logging.getLogger(__name__)


class NICAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        assert option_string in ('--port', '--network', '--ip', '--subnet')
        nics = getattr(namespace, self.dest, None) or []
        if option_string == '--ip':
            try:
                network, ip = values.split(':', 1)
            except ValueError:
                raise argparse.ArgumentError(
                    self, '--ip format is NETWORK:IP, got %s' % values)
            nics.append({'network': network, 'fixed_ip': ip})
        else:
            nics.append({option_string[2:]: values})
        setattr(namespace, self.dest, nics)


def _do_deploy(api, args, formatter):
    wait = None if args.no_wait else args.wait

    capabilities = dict(item.split('=', 1) for item in args.capability)
    if args.ssh_public_key:
        with open(args.ssh_public_key) as fp:
            ssh_keys = [fp.read().strip()]
    else:
        ssh_keys = []

    if args.hostname and not _utils.is_hostname_safe(args.hostname):
        raise RuntimeError("%s cannot be used as a hostname" % args.hostname)

    source = sources.detect(args.image,
                            kernel=args.image_kernel,
                            ramdisk=args.image_ramdisk,
                            checksum=args.image_checksum)

    config = _config.InstanceConfig(ssh_keys=ssh_keys)
    if args.user_name:
        config.add_user(args.user_name, sudo=args.passwordless_sudo)

    node = api.reserve_node(resource_class=args.resource_class,
                            conductor_group=args.conductor_group,
                            capabilities=capabilities,
                            traits=args.trait,
                            candidates=args.candidate)
    instance = api.provision_node(node,
                                  image=source,
                                  nics=args.nics,
                                  root_size_gb=args.root_size,
                                  swap_size_mb=args.swap_size,
                                  config=config,
                                  hostname=args.hostname,
                                  netboot=args.netboot,
                                  wait=wait)
    formatter.deploy(instance)


def _do_undeploy(api, args, formatter):
    node = api.unprovision_node(args.node, wait=args.wait)
    formatter.undeploy(node)


def _do_show(api, args, formatter):
    instances = api.show_instances(args.instance)
    formatter.show(instances)


def _do_wait(api, args, formatter):
    instances = api.wait_for_provisioning(args.instance,
                                          timeout=args.timeout)
    formatter.show(instances)


def _do_list(api, args, formatter):
    instances = api.list_instances()
    formatter.show(instances)


def _parse_args(args, config):
    parser = argparse.ArgumentParser(
        description='Deployment and Scheduling tool for Bare Metal')
    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument('-q', '--quiet', action='store_true',
                           help='output only errors')
    verbosity.add_argument('--debug', action='store_true',
                           help='output extensive logging')
    verbosity.add_argument('-v', '--verbose', action='count', default=0,
                           dest='verbosity',
                           help='increase output verbosity, can be specified '
                           'up to three times')
    parser.add_argument('--dry-run', action='store_true',
                        help='do not take any destructive actions')
    parser.add_argument('--format', choices=list(_format.FORMATS),
                        default=_format.DEFAULT_FORMAT,
                        help='output format')

    config.register_argparse_arguments(parser, sys.argv[1:])

    subparsers = parser.add_subparsers()

    deploy = subparsers.add_parser('deploy')
    deploy.set_defaults(func=_do_deploy)
    wait_grp = deploy.add_mutually_exclusive_group()
    wait_grp.add_argument('--wait', type=int, default=1800,
                          help='time (in seconds) to wait for node to become '
                          'active')
    wait_grp.add_argument('--no-wait', action='store_true',
                          help='disable waiting for deploy to finish')
    deploy.add_argument('--image', help='image to use (name, UUID or URL)',
                        required=True)
    deploy.add_argument('--image-checksum',
                        help='image MD5 checksum or URL with checksums')
    deploy.add_argument('--image-kernel', help='URL of the image\'s kernel')
    deploy.add_argument('--image-ramdisk', help='URL of the image\'s ramdisk')
    deploy.add_argument('--network', help='network to create a port on '
                        '(name or UUID)', dest='nics', action=NICAction)
    deploy.add_argument('--subnet', help='subnet to create a port on '
                        '(name or UUID)', dest='nics', action=NICAction)
    deploy.add_argument('--port', help='port to attach (name or UUID)',
                        dest='nics', action=NICAction)
    deploy.add_argument('--ip', help='attach IP from the network',
                        dest='nics', metavar='NETWORK:IP', action=NICAction)
    deploy.add_argument('--netboot', action='store_true',
                        help='boot from network instead of local disk')
    deploy.add_argument('--root-size', type=int,
                        help='root partition size (in GiB), defaults to '
                        '(local_gb - 1)')
    deploy.add_argument('--swap-size', type=int,
                        help='swap partition size (in MiB), defaults to '
                        'no swap')
    deploy.add_argument('--capability', action='append', metavar='NAME=VALUE',
                        default=[], help='capabilities the node should have')
    deploy.add_argument('--trait', action='append',
                        default=[], help='trait the node should have')
    deploy.add_argument('--ssh-public-key', help='SSH public key to load')
    deploy.add_argument('--hostname', help='Host name to use, defaults to '
                        'Node\'s name or UUID')
    deploy.add_argument('--resource-class',
                        help='node resource class to deploy')
    deploy.add_argument('--conductor-group',
                        help='conductor group to pick the node from')
    deploy.add_argument('--candidate', action='append',
                        help='A candidate node to use for scheduling (can be '
                        'specified several times)')
    deploy.add_argument('--user-name', help='Name of the admin user to create')
    deploy.add_argument('--passwordless-sudo', action='store_true',
                        help='allow password-less sudo for the user')

    undeploy = subparsers.add_parser('undeploy')
    undeploy.set_defaults(func=_do_undeploy)
    undeploy.add_argument('node', help='node UUID')
    undeploy.add_argument('--wait', type=int,
                          help='time (in seconds) to wait for node to become '
                          'available for deployment again')

    show = subparsers.add_parser('show')
    show.set_defaults(func=_do_show)
    show.add_argument('instance', nargs='+', help='instance UUID(s)')

    show = subparsers.add_parser('list')
    show.set_defaults(func=_do_list)

    wait = subparsers.add_parser('wait')
    wait.set_defaults(func=_do_wait)
    wait.add_argument('instance', nargs='+', help='instance UUID(s)')
    wait.add_argument('--timeout', type=int,
                      help='time (in seconds) to wait for provisioning.')

    return parser.parse_args(args)


_URLLIB3_LOGGER = 'urllib3.connectionpool'


def _configure_logging(args):
    log_fmt = ('%(asctime)s %(levelname)s %(name)s: %(message)s'
               if args.debug or args.verbosity
               else '[%(asctime)s] %(message)s')

    # Verbosity:
    # 0 (the default) - warnings and errors
    # 1 - info from metalsmith, warnings and errors from everything else
    # 2 - debug from metalsmith, info from everything else
    # 3 - the same as --debug
    base_level = logging.WARNING
    metalsmith_level = base_level
    urllib_level = logging.CRITICAL

    if args.quiet:
        base_level = logging.CRITICAL
        metalsmith_level = base_level
    elif args.debug or args.verbosity > 2:
        base_level = logging.DEBUG
        metalsmith_level = base_level
        urllib_level = logging.INFO
    elif args.verbosity == 2:
        base_level = logging.INFO
        metalsmith_level = logging.DEBUG
    elif args.verbosity == 1:
        metalsmith_level = logging.INFO

    logging.basicConfig(level=base_level, format=log_fmt)
    logging.getLogger('metalsmith').setLevel(metalsmith_level)
    logging.getLogger(_URLLIB3_LOGGER).setLevel(urllib_level)


def main(args=sys.argv[1:]):
    config = os_config.OpenStackConfig()
    args = _parse_args(args, config)
    _configure_logging(args)
    if args.quiet:
        formatter = _format.NULL_FORMAT
    else:
        formatter = _format.FORMATS[args.format]

    region = config.get_one(argparse=args)
    api = _provisioner.Provisioner(cloud_region=region, dry_run=args.dry_run)

    try:
        args.func(api, args, formatter)
    except Exception as exc:
        LOG.critical('%s', exc, exc_info=args.debug)
        sys.exit(1)
